from enum import Enum

from List.configDict import locationType, itemType
from List.inventory import magic, keyblade, ability, summon, report, storyunlock, form, misc
from List.location.graph import DefaultLogicGraph, RequirementEdge, chest, popup, hybrid_bonus, stat_bonus, item_bonus, \
    LocationGraphBuilder, START_NODE
from Module.itemPlacementRestriction import ItemPlacementHelpers


class NodeId(str, Enum):
    MarketplaceMapPopup = "Marketplace Map Popup"
    Borough = "Borough"
    BoroughChests = "Borough Chests"
    MerlinsHousePopup = "Merlin's House Popup"
    Bailey = "Bailey"
    BaseballCharmPopup = "Baseball Charm Popup"
    Postern = "Postern"
    PosternChests = "Postern Chests"
    Corridors = "Cooridors"
    CorridorsChests = "Cooridors Chests"
    DoorToDarknessPopup = "DTD Popup"
    AnsemsStudy = "Ansem's Study"
    AnsemsStudyChests = "Ansem's Study Chests"
    UkuleleCharm = "Ukulele Charm"
    UkuleleCharmChests = "Ukulele Charm Chests"
    RestorationSite = "Restoration Site"
    RestorationSiteChests = "Restoration Site Chests"
    DemyxHollowBastion = "Demyx (HB)"
    FinalFantasyFights = "Final Fantasy Fights"
    CrystalFissure = "Crystal Fissure"
    CrystalFissureChests = "Crystal Fissure Chests"
    ThousandHeartless = "1000 Heartless"
    GullWing = "Gull Wing"
    GullWingChests = "Gull Wing Chests"
    HeartlessManufactory = "Heartless Manufactory"
    HeartlessManufactoryChests = "Heartless Manufactory Chests"
    Sephiroth = "Sephiroth"
    Mushroom13 = "Mushroom 13"
    DataDemyx = "Data Demyx"
    CorDepths = "CoR Depths"
    CorDepthsChests = "CoR Depths Chests"
    CorMineshaftPreFigh1 = "CoR Mineshaft Pre-Fight 1"
    CorMineshaftPreFigh1Chests = "CoR Mineshaft Pre-Fight 1 Chests"
    CorDepthsPostFight1 = "CoR-Depths Post-Fight 1"
    CorDepthsPostFight1Chests = "CoR-Depths Post-Fight 1 Chests"
    CorMiningArea = "CoR Mining Area"
    CorMiningAreaChests = "CoR Mining Area Chests"
    CorMineshaftPreFight2 = "CoR Minshaft Pre-Fight 2"
    CorMineshaftPreFight2Chests = "CoR Minshaft Pre-Fight 2 Chests"
    CorEngineChamber = "CoR Engine Chamber"
    CorEngineChamberChests  = "CoR Engine Chamber Chests"
    CorMineshaftPostFight2 = "CoR Mineshaft Post Fight 2"
    CorMineshaftPostFight2Chests = "CoR Mineshaft Post Fight 2 Chests"
    CorMineshaftLastChest = "CoR Mineshaft Last Chest"
    CorMineshaftLastChestChests = "CoR Mineshaft Last Chest Chests"
    TransportToRemembrance = "Transport to Remembrance"


class CheckLocation(str, Enum):
    MarketplaceMap = "Marketplace Map"
    BoroughDriveRecovery = "Borough Drive Recovery"
    BoroughApBoost = "Borough AP Boost"
    BoroughHiPotion = "Borough Hi-Potion"
    BoroughMythrilShard = "Borough Mythril Shard"
    BoroughDarkShard = "Borough Dark Shard"
    MerlinsHouseMembershipCard = "Merlin's House Membership Card"
    MerlinsHouseBlizzardElement = "Merlin's House Blizzard Element"
    Bailey = "Bailey"
    BaileySecretAnsemReport7 = "Bailey Secret Ansem Report 7"
    BaseballCharm = "Baseball Charm"
    PosternCastlePerimeterMap = "Postern Castle Perimeter Map"
    PosternMythrilGem = "Postern Mythril Gem"
    PosternApBoost = "Postern AP Boost"
    CorridorsMythrilStone = "Corridors Mythril Stone"
    CorridorsMythrilCrystal = "Corridors Mythril Crystal"
    CorridorsDarkCrystal = "Corridors Dark Crystal"
    CorridorsApBoost = "Corridors AP Boost"
    AnsemsStudyMasterForm = "Ansem's Study Master Form"
    AnsemsStudySleepingLion = "Ansem's Study Sleeping Lion"
    AnsemsStudySkillRecipe = "Ansem's Study Skill Recipe"
    UkuleleCharm = "Ansem's Study Ukulele Charm"
    RestorationSiteMoonRecipe = "Restoration Site Moon Recipe"
    RestorationSiteApBoost = "Restoration Site AP Boost"
    DemyxHollowBastion = "Demyx (HB)"
    FinalFantasyFightsCureElement = "FF Fights Cure Element"
    CrystalFissureTornPages = "Crystal Fissure Torn Pages"
    CrystalFissureGreatMawMap = "Crystal Fissure The Great Maw Map"
    CrystalFissureEnergyCrystal = "Crystal Fissure Energy Crystal"
    CrystalFissureApBoost = "Crystal Fissure AP Boost"
    ThousandHeartless = "1000 Heartless"
    ThousandHeartlessSecretAnsemReport1 = "1000 Heartless Secret Ansem Report 1"
    ThousandHeartlessIceCream = "1000 Heartless Ice Cream"
    ThousandHeartlessPicture = "1000 Heartless Picture"
    PosternGullWing = "Postern Gull Wing"
    HeartlessManufactoryCosmicChain = "Heartless Manufactory Cosmic Chain"
    SephirothBonus = "Sephiroth Bonus"
    SephirothFenrir = "Sephiroth Fenrir"
    WinnersProof = "Winner's Proof"
    ProofOfPeace = "Proof of Peace"
    DataDemyxApBoost = "Demyx (Data) AP Boost"

    CorDepthsApBoost = "CoR Depths AP Boost"
    CorDepthsPowerCrystal = "CoR Depths Power Crystal"
    CorDepthsFrostCrystal = "CoR Depths Frost Crystal"
    CorDepthsManifestIllusion = "CoR Depths Manifest Illusion"
    CorDepthsApBoost2 = "CoR Depths AP Boost (2)"
    CorMineshaftLowerLevelDepthsOfRemembranceMap = "CoR Mineshaft Lower Level Depths of Remembrance Map"
    CorMineshaftLowerLevelApBoost = "CoR Mineshaft Lower Level AP Boost"
    CorDepthsUpperLevelRemembranceGem = "CoR Depths Upper Level Remembrance Gem"
    CorMiningAreaSerenityGem = "CoR Mining Area Serenity Gem"
    CorMiningAreaApBoost = "CoR Mining Area AP Boost"
    CorMiningAreaSerenityCrystal = "CoR Mining Area Serenity Crystal"
    CorMiningAreaManifestIllusion = "CoR Mining Area Manifest Illusion"
    CorMiningAreaSerenityGem2 = "CoR Mining Area Serenity Gem (2)"
    CorMiningAreaDarkRemembranceMap = "CoR Mining Area Dark Remembrance Map"
    CorMineshaftMidLevelPowerBoost = "CoR Mineshaft Mid Level Power Boost"
    CorEngineChamberSerenityCrystal = "CoR Engine Chamber Serenity Crystal"
    CorEngineChamberRemembranceCrystal = "CoR Engine Chamber Remembrance Crystal"
    CorEngineChamberApBoost = "CoR Engine Chamber AP Boost"
    CorEngineChamberManifestIllusion = "CoR Engine Chamber Manifest Illusion"
    CorMineshaftUpperLevelMagicBoost = "CoR Mineshaft Upper Level Magic Boost"
    CorMineshaftUpperLevelApBoost = "CoR Mineshaft Upper Level AP Boost"
    TransportToRemembrance = "Transport to Remembrance"

class HBLogicGraph(DefaultLogicGraph):
    def __init__(self,reverse_rando,keyblade_unlocks):
        DefaultLogicGraph.__init__(self,NodeId)
        hb_keyblade_lambda = lambda inv : not keyblade_unlocks or ItemPlacementHelpers.need_hb_keyblade(inv)
        cor_keyblade_lambda = lambda inv : not keyblade_unlocks or ItemPlacementHelpers.need_cor_keyblade(inv)
        self.logic[NodeId.Borough][NodeId.BoroughChests] = hb_keyblade_lambda
        self.logic[NodeId.Postern][NodeId.PosternChests] = hb_keyblade_lambda
        self.logic[NodeId.Corridors][NodeId.CorridorsChests] = hb_keyblade_lambda
        self.logic[NodeId.AnsemsStudy][NodeId.AnsemsStudyChests] = hb_keyblade_lambda
        self.logic[NodeId.UkuleleCharm][NodeId.UkuleleCharmChests] = hb_keyblade_lambda
        self.logic[NodeId.RestorationSite][NodeId.RestorationSiteChests] = hb_keyblade_lambda
        self.logic[NodeId.CrystalFissure][NodeId.CrystalFissureChests] = hb_keyblade_lambda
        self.logic[NodeId.GullWing][NodeId.GullWingChests] = hb_keyblade_lambda
        self.logic[NodeId.HeartlessManufactory][NodeId.HeartlessManufactoryChests] = hb_keyblade_lambda

        self.logic[NodeId.CorDepths][NodeId.CorDepthsChests] = cor_keyblade_lambda
        self.logic[NodeId.CorMineshaftPreFigh1][NodeId.CorMineshaftPreFigh1Chests] = cor_keyblade_lambda
        self.logic[NodeId.CorDepthsPostFight1][NodeId.CorDepthsPostFight1Chests] = cor_keyblade_lambda
        self.logic[NodeId.CorMiningArea][NodeId.CorMiningAreaChests] = cor_keyblade_lambda
        self.logic[NodeId.CorMineshaftPreFight2][NodeId.CorMineshaftPreFight2Chests] = cor_keyblade_lambda
        self.logic[NodeId.CorEngineChamber][NodeId.CorEngineChamberChests] = cor_keyblade_lambda
        self.logic[NodeId.CorMineshaftPostFight2][NodeId.CorMineshaftPostFight2Chests] = cor_keyblade_lambda
        self.logic[NodeId.CorMineshaftLastChest][NodeId.CorMineshaftLastChestChests] = cor_keyblade_lambda
        if not reverse_rando:
            self.logic[START_NODE][NodeId.MarketplaceMapPopup] = ItemPlacementHelpers.hb1_check
            self.logic[NodeId.BaseballCharmPopup][NodeId.Postern] = ItemPlacementHelpers.hb2_check
            self.logic[NodeId.DoorToDarknessPopup][NodeId.CorDepths] = ItemPlacementHelpers.need_growths
            self.logic[NodeId.ThousandHeartless][NodeId.Mushroom13] = ItemPlacementHelpers.need_proof_peace
            self.logic[NodeId.ThousandHeartless][NodeId.DataDemyx] = ItemPlacementHelpers.need_forms
        else:
            self.logic[START_NODE][NodeId.Borough] = ItemPlacementHelpers.hb1_check
            self.logic[NodeId.ThousandHeartless][NodeId.CorDepths] = lambda inv : ItemPlacementHelpers.hb2_check(inv) and ItemPlacementHelpers.need_growths(inv)
            self.logic[NodeId.BaseballCharmPopup][NodeId.DataDemyx] = ItemPlacementHelpers.need_forms
            

def make_graph(graph: LocationGraphBuilder):
    hb = locationType.HB
    cor = locationType.CoR
    hb_logic = HBLogicGraph(graph.reverse_rando,graph.keyblades_unlock_chests)
    graph.add_logic(hb_logic)

    marketplace_map_popup = graph.add_location(NodeId.MarketplaceMapPopup, [
        popup(362, CheckLocation.MarketplaceMap, hb),
    ])
    borough_chests = graph.add_location(NodeId.BoroughChests, [
        chest(194, CheckLocation.BoroughDriveRecovery, hb),
        chest(195, CheckLocation.BoroughApBoost, hb),
        chest(196, CheckLocation.BoroughHiPotion, hb),
        chest(305, CheckLocation.BoroughMythrilShard, hb),
        chest(506, CheckLocation.BoroughDarkShard, hb),
    ])
    borough = graph.add_location(NodeId.Borough, [])
    merlins_house_popup = graph.add_location(NodeId.MerlinsHousePopup, [
        popup(256, CheckLocation.MerlinsHouseMembershipCard, hb, vanilla=storyunlock.MembershipCard),
        popup(292, CheckLocation.MerlinsHouseBlizzardElement, hb, vanilla=magic.Blizzard),
    ])
    bailey = graph.add_location(NodeId.Bailey, [
        item_bonus(47, CheckLocation.Bailey, hb, vanilla=magic.Fire),
        popup(531, CheckLocation.BaileySecretAnsemReport7, hb, vanilla=report.AnsemReport7),
    ])
    baseball_charm_popup = graph.add_location(NodeId.BaseballCharmPopup, [
        popup(258, CheckLocation.BaseballCharm, hb, vanilla=summon.BaseballCharm),
    ])
    postern_chests = graph.add_location(NodeId.PosternChests, [
        chest(310, CheckLocation.PosternCastlePerimeterMap, hb),
        chest(189, CheckLocation.PosternMythrilGem, hb),
        chest(190, CheckLocation.PosternApBoost, hb),
    ])
    postern = graph.add_location(NodeId.Postern, [])
    corridors_chests = graph.add_location(NodeId.CorridorsChests, [
        chest(200, CheckLocation.CorridorsMythrilStone, hb),
        chest(201, CheckLocation.CorridorsMythrilCrystal, hb),
        chest(202, CheckLocation.CorridorsDarkCrystal, hb),
        chest(307, CheckLocation.CorridorsApBoost, hb),
    ])
    corridors = graph.add_location(NodeId.Corridors, [])
    dtd_popup = graph.add_location(NodeId.DoorToDarknessPopup, [
        popup(266, CheckLocation.AnsemsStudyMasterForm, hb, vanilla=form.MasterForm),
        popup(276, CheckLocation.AnsemsStudySleepingLion, hb, vanilla=keyblade.SleepingLion),
    ])
    ansems_study_chests = graph.add_location(NodeId.AnsemsStudyChests, [
        chest(184, CheckLocation.AnsemsStudySkillRecipe, hb),
    ])
    ansems_study = graph.add_location(NodeId.AnsemsStudy, [])
    ukulele_charm_chests = graph.add_location(NodeId.UkuleleCharmChests, [
        chest(183, CheckLocation.UkuleleCharm, hb, vanilla=summon.UkuleleCharm),
    ])
    ukulele_charm = graph.add_location(NodeId.UkuleleCharm, [])
    restoration_site_chests = graph.add_location(NodeId.RestorationSiteChests, [
        chest(309, CheckLocation.RestorationSiteMoonRecipe, hb),
        chest(507, CheckLocation.RestorationSiteApBoost, hb),
    ])
    restoration_site = graph.add_location(NodeId.RestorationSite, [])
    demyx = graph.add_location(NodeId.DemyxHollowBastion, [
        hybrid_bonus(28, CheckLocation.DemyxHollowBastion, hb, vanilla=magic.Blizzard),
    ])
    final_fantasy_fights = graph.add_location(NodeId.FinalFantasyFights, [
        popup(361, CheckLocation.FinalFantasyFightsCureElement, hb, vanilla=magic.Cure),
    ])
    crystal_fissure_chests = graph.add_location(NodeId.CrystalFissureChests, [
        chest(179, CheckLocation.CrystalFissureTornPages, hb, vanilla=misc.TornPages),
        chest(489, CheckLocation.CrystalFissureGreatMawMap, hb),
        chest(180, CheckLocation.CrystalFissureEnergyCrystal, hb),
        chest(181, CheckLocation.CrystalFissureApBoost, hb),
    ])
    crystal_fissure = graph.add_location(NodeId.CrystalFissure, [])
    thousand_heartless = graph.add_location(NodeId.ThousandHeartless, [
        item_bonus(60, CheckLocation.ThousandHeartless, hb, vanilla=ability.GuardBreak),
        popup(525, CheckLocation.ThousandHeartlessSecretAnsemReport1, hb, vanilla=report.AnsemReport1),
        popup(269, CheckLocation.ThousandHeartlessIceCream, hb, vanilla=storyunlock.IceCream),
        popup(511, CheckLocation.ThousandHeartlessPicture, hb),
    ])
    gull_wing_chests = graph.add_location(NodeId.GullWingChests, [
        chest(491, CheckLocation.PosternGullWing, hb, vanilla=keyblade.GullWing),
    ])
    gull_wing = graph.add_location(NodeId.GullWing, [])
    heartless_manufactory_chests = graph.add_location(NodeId.HeartlessManufactoryChests, [
        chest(311, CheckLocation.HeartlessManufactoryCosmicChain, hb, invalid_checks=[itemType.MANUFACTORYUNLOCK]),
    ])
    heartless_manufactory = graph.add_location(NodeId.HeartlessManufactory, [])
    sephiroth = graph.add_location(NodeId.Sephiroth, [
        stat_bonus(35, CheckLocation.SephirothBonus, [hb, locationType.Sephi]),
        popup(282, CheckLocation.SephirothFenrir, [hb, locationType.Sephi], vanilla=keyblade.Fenrir),
    ])
    mushroom_13 = graph.add_location(NodeId.Mushroom13, [
        popup(588, CheckLocation.WinnersProof, [hb, locationType.Mush13], invalid_checks=[itemType.PROOF_OF_PEACE],
              vanilla=keyblade.WinnersProof),
        popup(589, CheckLocation.ProofOfPeace, [hb, locationType.Mush13], invalid_checks=[itemType.PROOF_OF_PEACE]),
    ])
    data_demyx = graph.add_location(NodeId.DataDemyx, [
        popup(560, CheckLocation.DataDemyxApBoost, [hb, locationType.DataOrg], invalid_checks=[itemType.FORM]),
    ])

    cor_depths_chests = graph.add_location(NodeId.CorDepthsChests, [
        chest(562, CheckLocation.CorDepthsApBoost, [hb, cor]),
        chest(563, CheckLocation.CorDepthsPowerCrystal, [hb, cor]),
        chest(564, CheckLocation.CorDepthsFrostCrystal, [hb, cor]),
        chest(565, CheckLocation.CorDepthsManifestIllusion, [hb, cor]),
        chest(566, CheckLocation.CorDepthsApBoost2, [hb, cor]),
    ])
    cor_depths = graph.add_location(NodeId.CorDepths, [])
    cor_mineshaft_pre_fight_1_chests = graph.add_location(NodeId.CorMineshaftPreFigh1Chests, [
        chest(580, CheckLocation.CorMineshaftLowerLevelDepthsOfRemembranceMap, [hb, cor]),
        chest(578, CheckLocation.CorMineshaftLowerLevelApBoost, [hb, cor]),
    ])
    cor_mineshaft_pre_fight_1 = graph.add_location(NodeId.CorMineshaftPreFigh1, [])
    cor_depths_post_fight_1_chests = graph.add_location(NodeId.CorDepthsPostFight1Chests, [
        chest(567, CheckLocation.CorDepthsUpperLevelRemembranceGem, [hb, cor]),
    ])
    cor_depths_post_fight_1 = graph.add_location(NodeId.CorDepthsPostFight1, [])
    cor_mining_area_chests = graph.add_location(NodeId.CorMiningAreaChests, [
        chest(568, CheckLocation.CorMiningAreaSerenityGem, [hb, cor]),
        chest(569, CheckLocation.CorMiningAreaApBoost, [hb, cor]),
        chest(570, CheckLocation.CorMiningAreaSerenityCrystal, [hb, cor]),
        chest(571, CheckLocation.CorMiningAreaManifestIllusion, [hb, cor]),
        chest(572, CheckLocation.CorMiningAreaSerenityGem2, [hb, cor]),
        chest(573, CheckLocation.CorMiningAreaDarkRemembranceMap, [hb, cor]),
    ])
    cor_mining_area = graph.add_location(NodeId.CorMiningArea, [])
    cor_mineshaft_pre_fight_2_chests = graph.add_location(NodeId.CorMineshaftPreFight2Chests, [
        chest(581, CheckLocation.CorMineshaftMidLevelPowerBoost, [hb, cor]),
    ])
    cor_mineshaft_pre_fight_2 = graph.add_location(NodeId.CorMineshaftPreFight2, [])
    cor_engine_chamber_chests = graph.add_location(NodeId.CorEngineChamberChests, [
        chest(574, CheckLocation.CorEngineChamberSerenityCrystal, [hb, cor]),
        chest(575, CheckLocation.CorEngineChamberRemembranceCrystal, [hb, cor]),
        chest(576, CheckLocation.CorEngineChamberApBoost, [hb, cor]),
        chest(577, CheckLocation.CorEngineChamberManifestIllusion, [hb, cor]),
    ])
    cor_engine_chamber = graph.add_location(NodeId.CorEngineChamber, [])
    cor_mineshaft_post_fight_2_chests = graph.add_location(NodeId.CorMineshaftPostFight2Chests, [
        chest(582, CheckLocation.CorMineshaftUpperLevelMagicBoost, [hb, cor]),
    ])
    cor_mineshaft_post_fight_2 = graph.add_location(NodeId.CorMineshaftPostFight2, [])
    cor_mineshaft_last_chest_chests = graph.add_location(NodeId.CorMineshaftLastChestChests, [
        chest(579, CheckLocation.CorMineshaftUpperLevelApBoost, [hb, cor]),
    ])
    cor_mineshaft_last_chest = graph.add_location(NodeId.CorMineshaftLastChest, [])
    transport_to_remembrance = graph.add_location(NodeId.TransportToRemembrance, [
        stat_bonus(72, CheckLocation.TransportToRemembrance, [hb, locationType.TTR]),
    ])

    graph.register_superboss(data_demyx)
    
    graph.add_edge(borough, borough_chests)
    graph.add_edge(postern, postern_chests)
    graph.add_edge(corridors, corridors_chests)
    graph.add_edge(ansems_study, ansems_study_chests)
    graph.add_edge(ukulele_charm, ukulele_charm_chests)
    graph.add_edge(restoration_site, restoration_site_chests)
    graph.add_edge(crystal_fissure, crystal_fissure_chests)
    graph.add_edge(gull_wing, gull_wing_chests)
    graph.add_edge(heartless_manufactory, heartless_manufactory_chests)
    
    graph.add_edge(cor_depths, cor_depths_chests)
    graph.add_edge(cor_mineshaft_pre_fight_1, cor_mineshaft_pre_fight_1_chests)
    graph.add_edge(cor_depths_post_fight_1, cor_depths_post_fight_1_chests)
    graph.add_edge(cor_mining_area, cor_mining_area_chests)
    graph.add_edge(cor_mineshaft_pre_fight_2, cor_mineshaft_pre_fight_2_chests)
    graph.add_edge(cor_engine_chamber, cor_engine_chamber_chests)
    graph.add_edge(cor_mineshaft_post_fight_2, cor_mineshaft_post_fight_2_chests)
    graph.add_edge(cor_mineshaft_last_chest, cor_mineshaft_last_chest_chests)

    if not graph.reverse_rando:
        graph.add_edge(START_NODE, marketplace_map_popup)
        graph.add_edge(marketplace_map_popup, borough)
        graph.add_edge(borough, merlins_house_popup)
        graph.add_edge(merlins_house_popup, bailey, RequirementEdge(battle=True))
        graph.add_edge(bailey, baseball_charm_popup)
        graph.add_edge(baseball_charm_popup, postern)
        graph.add_edge(postern, corridors)
        graph.add_edge(corridors, dtd_popup)
        graph.add_edge(corridors, ansems_study)
        graph.add_edge(dtd_popup, ukulele_charm)

        graph.add_edge(dtd_popup, cor_depths)
        graph.add_edge(cor_depths, cor_mineshaft_pre_fight_1)
        graph.add_edge(cor_mineshaft_pre_fight_1, cor_depths_post_fight_1, RequirementEdge(battle=True))
        graph.add_edge(cor_depths_post_fight_1, cor_mining_area)
        graph.add_edge(cor_mining_area, cor_mineshaft_pre_fight_2)
        graph.add_edge(cor_mineshaft_pre_fight_2, cor_engine_chamber, RequirementEdge(battle=True))
        graph.add_edge(cor_engine_chamber, cor_mineshaft_post_fight_2)
        graph.add_edge(cor_mineshaft_post_fight_2, cor_mineshaft_last_chest)
        graph.add_edge(cor_mineshaft_last_chest, transport_to_remembrance, RequirementEdge(battle=True))

        graph.add_edge(dtd_popup, restoration_site, RequirementEdge(battle=True))
        graph.add_edge(restoration_site, demyx, RequirementEdge(battle=True))
        graph.add_edge(demyx, final_fantasy_fights, RequirementEdge(battle=True))
        graph.add_edge(final_fantasy_fights, crystal_fissure)
        graph.add_edge(crystal_fissure, thousand_heartless, RequirementEdge(battle=True))
        graph.add_edge(thousand_heartless, gull_wing)
        graph.add_edge(thousand_heartless, heartless_manufactory)
        graph.add_edge(thousand_heartless, sephiroth, RequirementEdge(battle=True))
        graph.add_edge(thousand_heartless, mushroom_13)
        graph.add_edge(thousand_heartless, data_demyx,
                       RequirementEdge(battle=True))
        graph.register_first_boss(baseball_charm_popup)
        graph.register_last_story_boss(thousand_heartless)
        graph.register_superboss(sephiroth)
    else:
        graph.add_edge(START_NODE, borough)
        graph.add_edge(borough, postern)
        graph.add_edge(postern, corridors)
        graph.add_edge(corridors, dtd_popup)
        graph.add_edge(corridors, ansems_study)
        graph.add_edge(dtd_popup, ukulele_charm)
        graph.add_edge(dtd_popup, restoration_site, RequirementEdge(battle=True))
        graph.add_edge(restoration_site, demyx, RequirementEdge(battle=True))
        graph.add_edge(demyx, final_fantasy_fights, RequirementEdge(battle=True))
        graph.add_edge(final_fantasy_fights, crystal_fissure)
        graph.add_edge(crystal_fissure, thousand_heartless, RequirementEdge(battle=True))
        graph.add_edge(thousand_heartless, gull_wing)
        graph.add_edge(thousand_heartless, heartless_manufactory)
        graph.add_edge(thousand_heartless, cor_depths)
        graph.add_edge(cor_depths, cor_mineshaft_pre_fight_1)
        graph.add_edge(cor_mineshaft_pre_fight_1, cor_depths_post_fight_1, RequirementEdge(battle=True))
        graph.add_edge(cor_depths_post_fight_1, cor_mining_area)
        graph.add_edge(cor_mining_area, cor_mineshaft_pre_fight_2)
        graph.add_edge(cor_mineshaft_pre_fight_2, cor_engine_chamber, RequirementEdge(battle=True))
        graph.add_edge(cor_engine_chamber, cor_mineshaft_post_fight_2)
        graph.add_edge(cor_mineshaft_post_fight_2, cor_mineshaft_last_chest)
        graph.add_edge(cor_mineshaft_last_chest, transport_to_remembrance, RequirementEdge(battle=True))
        graph.add_edge(transport_to_remembrance, mushroom_13)
        graph.add_edge(transport_to_remembrance, sephiroth)
        graph.add_edge(sephiroth, marketplace_map_popup)
        graph.add_edge(marketplace_map_popup, merlins_house_popup)
        graph.add_edge(merlins_house_popup, bailey, RequirementEdge(battle=True))
        graph.add_edge(bailey, baseball_charm_popup)
        graph.add_edge(baseball_charm_popup, data_demyx,
                       RequirementEdge(battle=True))
        graph.register_first_boss(thousand_heartless)
        graph.register_last_story_boss(bailey)



from enum import Enum

from List.configDict import locationType
from List.inventory import magic, keyblade, misc
from List.location.graph import DefaultLogicGraph, RequirementEdge, chest, popup, hybrid_bonus, stat_bonus, LocationGraphBuilder, \
    START_NODE
from Module.itemPlacementRestriction import ItemPlacementHelpers


class NodeId(str, Enum):
    Gorge = "Gorge"
    GorgeChests = "Gorge Chests"
    ElephantGraveyard = "Elephant Graveyard"
    ElephantGraveyardChests = "Elephant Graveyard Chests"
    PrideRock = "Pride Rock"
    PrideRockChests = "Pride Rock Chests"
    WildebeestValley = "Wildebeest Valley"
    WildebeestValleyChests = "Wildebeest Valley Chests"
    Wastelands = "Wastelands"
    WastelandsChests = "Wastelands Chests"
    Jungle = "Jungle"
    JungleChests = "Jungle Chests"
    Oasis = "Oasis"
    OasisChests = "Oasis Chests"
    CircleOfLife = "Circle of Life"
    Hyenas1Bonus = "Hyenas 1 Bonus"
    Scar = "Scar"
    Hyenas2Bonus = "Hyenas 2 Bonus"
    Groundshaker = "Groundshaker"
    DataSaix = "Data Saix"


class CheckLocation(str, Enum):
    GorgeSavannahMap = "Gorge Savannah Map"
    GorgeDarkGem = "Gorge Dark Gem"
    GorgeMyhtrilStone = "Gorge Mythril Stone"
    ElephantGraveyardFrostGem = "Elephant Graveyard Frost Gem"
    ElephantGraveyardMythrilStone = "Elephant Graveyard Mythril Stone"
    ElephantGraveyardBrightStone = "Elephant Graveyard Bright Stone"
    ElephantGraveyardApBoost = "Elephant Graveyard AP Boost"
    ElephantGraveyardMythrilShard = "Elephant Graveyard Mythril Shard"
    PrideRockMap = "Pride Rock Map"
    PrideRockMythrilStone = "Pride Rock Mythril Stone"
    PrideRockSerenityCrystal = "Pride Rock Serenity Crystal"
    WildebeestValleyEnergyStone = "Wildebeest Valley Energy Stone"
    WildebeestValleyApBoost = "Wildebeest Valley AP Boost"
    WildebeestValleyMythrilGem = "Wildebeest Valley Mythril Gem"
    WildebeestValleyMythrilStone = "Wildebeest Valley Mythril Stone"
    WildebeestValleyLucidGem = "Wildebeest Valley Lucid Gem"
    WastelandsMythrilShard = "Wastelands Mythril Shard"
    WastelandsSerenityGem = "Wastelands Serenity Gem"
    WastelandsMythrilStone = "Wastelands Mythril Stone"
    JungleSerenityGem = "Jungle Serenity Gem"
    JungleMythrilStone = "Jungle Mythril Stone"
    JungleSerenityCrystal = "Jungle Serenity Crystal"
    OasisMap = "Oasis Map"
    OasisTornPages = "Oasis Torn Pages"
    OasisApBoost = "Oasis AP Boost"
    CircleOfLife = "Circle of Life"
    Hyenas1 = "Hyenas 1"
    Scar = "Scar"
    ScarFireElement = "Scar Fire Element"
    Hyenas2 = "Hyenas 2"
    Groundshaker = "Groundshaker"
    DataSaix = "Saix (Data) Defense Boost"

class PLLogicGraph(DefaultLogicGraph):
    def __init__(self,reverse_rando,keyblade_unlocks):
        DefaultLogicGraph.__init__(self,NodeId)
        keyblade_lambda = lambda inv : not keyblade_unlocks or ItemPlacementHelpers.need_pl_keyblade(inv)
        self.logic[NodeId.Gorge][NodeId.GorgeChests] = keyblade_lambda
        self.logic[NodeId.ElephantGraveyard][NodeId.ElephantGraveyardChests] = keyblade_lambda
        self.logic[NodeId.PrideRock][NodeId.PrideRockChests] = keyblade_lambda
        self.logic[NodeId.WildebeestValley][NodeId.WildebeestValleyChests] = keyblade_lambda
        self.logic[NodeId.Wastelands][NodeId.WastelandsChests] = keyblade_lambda
        self.logic[NodeId.Jungle][NodeId.JungleChests] = keyblade_lambda
        self.logic[NodeId.Oasis][NodeId.OasisChests] = keyblade_lambda
        if not reverse_rando:
            self.logic[START_NODE][NodeId.Gorge] = ItemPlacementHelpers.pl1_check
            self.logic[NodeId.Scar][NodeId.Hyenas2Bonus] = ItemPlacementHelpers.pl2_check
        else:
            self.logic[START_NODE][NodeId.PrideRock] = ItemPlacementHelpers.pl1_check
            self.logic[NodeId.Groundshaker][NodeId.Gorge] = ItemPlacementHelpers.pl2_check
           

def make_graph(graph: LocationGraphBuilder):
    pl = locationType.PL
    pl_logic = PLLogicGraph(graph.reverse_rando,graph.keyblades_unlock_chests)
    graph.add_logic(pl_logic)

    gorge_chests = graph.add_location(NodeId.GorgeChests, [
        chest(492, CheckLocation.GorgeSavannahMap, pl),
        chest(404, CheckLocation.GorgeDarkGem, pl),
        chest(405, CheckLocation.GorgeMyhtrilStone, pl),
    ])
    gorge = graph.add_location(NodeId.Gorge, [])
    elephant_graveyard_chests = graph.add_location(NodeId.ElephantGraveyardChests, [
        chest(401, CheckLocation.ElephantGraveyardFrostGem, pl),
        chest(402, CheckLocation.ElephantGraveyardMythrilStone, pl),
        chest(403, CheckLocation.ElephantGraveyardBrightStone, pl),
        chest(508, CheckLocation.ElephantGraveyardApBoost, pl),
        chest(509, CheckLocation.ElephantGraveyardMythrilShard, pl),
    ])
    elephant_graveyard = graph.add_location(NodeId.ElephantGraveyard, [])
    pride_rock_chests = graph.add_location(NodeId.PrideRockChests, [
        chest(418, CheckLocation.PrideRockMap, pl),
        chest(392, CheckLocation.PrideRockMythrilStone, pl),
        chest(393, CheckLocation.PrideRockSerenityCrystal, pl),
    ])
    pride_rock = graph.add_location(NodeId.PrideRock, [])
    wildebeest_valley_chests = graph.add_location(NodeId.WildebeestValleyChests, [
        chest(396, CheckLocation.WildebeestValleyEnergyStone, pl),
        chest(397, CheckLocation.WildebeestValleyApBoost, pl),
        chest(398, CheckLocation.WildebeestValleyMythrilGem, pl),
        chest(399, CheckLocation.WildebeestValleyMythrilStone, pl),
        chest(400, CheckLocation.WildebeestValleyLucidGem, pl),
    ])
    wildebeest_valley = graph.add_location(NodeId.WildebeestValley, [])
    wastelands_chests = graph.add_location(NodeId.WastelandsChests, [
        chest(406, CheckLocation.WastelandsMythrilShard, pl),
        chest(407, CheckLocation.WastelandsSerenityGem, pl),
        chest(408, CheckLocation.WastelandsMythrilStone, pl),
    ])
    wastelands = graph.add_location(NodeId.Wastelands, [])
    jungle_chests = graph.add_location(NodeId.JungleChests, [
        chest(409, CheckLocation.JungleSerenityGem, pl),
        chest(410, CheckLocation.JungleMythrilStone, pl),
        chest(411, CheckLocation.JungleSerenityCrystal, pl),
    ])
    jungle = graph.add_location(NodeId.Jungle, [])
    oasis_chests = graph.add_location(NodeId.OasisChests, [
        chest(412, CheckLocation.OasisMap, pl),
        chest(493, CheckLocation.OasisTornPages, pl, vanilla=misc.TornPages),
        chest(413, CheckLocation.OasisApBoost, pl),
    ])
    oasis = graph.add_location(NodeId.Oasis, [])
    circle_of_life = graph.add_location(NodeId.CircleOfLife, [
        popup(264, CheckLocation.CircleOfLife, pl, vanilla=keyblade.CircleOfLife),
    ])
    hyenas_1_bonus = graph.add_location(NodeId.Hyenas1Bonus, [
        stat_bonus(49, CheckLocation.Hyenas1, pl),
    ])
    scar = graph.add_location(NodeId.Scar, [
        stat_bonus(29, CheckLocation.Scar, pl),
        popup(302, CheckLocation.ScarFireElement, pl, vanilla=magic.Fire),
    ])
    hyenas_2_bonus = graph.add_location(NodeId.Hyenas2Bonus, [
        stat_bonus(50, CheckLocation.Hyenas2, pl),
    ])
    groundshaker = graph.add_location(NodeId.Groundshaker, [
        hybrid_bonus(30, CheckLocation.Groundshaker, pl, vanilla=magic.Thunder),
    ])
    data_saix = graph.add_location(NodeId.DataSaix, [
        popup(556, CheckLocation.DataSaix, [pl, locationType.DataOrg]),
    ])

    graph.register_superboss(data_saix)
    
    graph.add_edge(gorge, gorge_chests)
    graph.add_edge(elephant_graveyard, elephant_graveyard_chests)
    graph.add_edge(pride_rock, pride_rock_chests)
    graph.add_edge(wildebeest_valley, wildebeest_valley_chests)
    graph.add_edge(wastelands, wastelands_chests)
    graph.add_edge(jungle, jungle_chests)
    graph.add_edge(oasis, oasis_chests)

    if not graph.reverse_rando:
        graph.add_edge(START_NODE, gorge)
        graph.add_edge(gorge, elephant_graveyard, RequirementEdge(battle=True))
        graph.add_edge(elephant_graveyard, pride_rock)
        graph.add_edge(pride_rock, wildebeest_valley)
        graph.add_edge(wildebeest_valley, wastelands)
        graph.add_edge(wastelands, jungle)
        graph.add_edge(jungle, oasis)
        graph.add_edge(oasis, circle_of_life)
        graph.add_edge(circle_of_life, hyenas_1_bonus, RequirementEdge(battle=True))
        graph.add_edge(hyenas_1_bonus, scar, RequirementEdge(battle=True))
        graph.add_edge(scar, hyenas_2_bonus, RequirementEdge(battle=True, req=ItemPlacementHelpers.pl2_check))
        graph.add_edge(hyenas_2_bonus, groundshaker, RequirementEdge(battle=True))
        graph.add_edge(groundshaker, data_saix, RequirementEdge(battle=True))
        graph.register_first_boss(scar)
        graph.register_last_story_boss(groundshaker)
    else:
        graph.add_edge(START_NODE, pride_rock)
        graph.add_edge(pride_rock, elephant_graveyard)
        graph.add_edge(elephant_graveyard, hyenas_2_bonus, RequirementEdge(battle=True))
        graph.add_edge(hyenas_2_bonus, wildebeest_valley)
        graph.add_edge(wildebeest_valley, wastelands)
        graph.add_edge(wastelands, jungle)
        graph.add_edge(jungle, groundshaker, RequirementEdge(battle=True))
        graph.add_edge(groundshaker, gorge, RequirementEdge(req=ItemPlacementHelpers.pl2_check))
        graph.add_edge(gorge, oasis, RequirementEdge(battle=True))
        graph.add_edge(oasis, circle_of_life)
        graph.add_edge(circle_of_life, hyenas_1_bonus, RequirementEdge(battle=True))
        graph.add_edge(hyenas_1_bonus, scar, RequirementEdge(battle=True))
        graph.add_edge(scar, data_saix, RequirementEdge(battle=True))
        graph.register_first_boss(groundshaker)
        graph.register_last_story_boss(scar)



from math import ceil, floor
from typing import Union

from Class.newLocationClass import KH2Location
from List.LvupStats import DreamWeaponOffsets
from List.NewLocationList import Locations
from List.configDict import itemType, locationCategory, itemRarity, locationDepth, locationType, itemBias, itemDifficulty
from List.location.graph import START_NODE
from Module.RandomizerSettings import RandomizerSettings




class SimplifiedWeightDistributions:
    def __init__(self, max_depth: int):
        self.weighting_function: dict[itemBias, list[int]] = {}

        weight_list_maker = lambda function_index : [
                    max(1, floor(10.0 / pow((max_depth + 2 - x) / (max_depth / 3.0), function_index * .5 + 1)))
                    for x in range(1, max_depth + 2)
                ]

        self.weighting_function[itemBias.VERY_EARLY]  = weight_list_maker(1)[::-1]
        self.weighting_function[itemBias.EARLY]  = weight_list_maker(0)[::-1]
        self.weighting_function[itemBias.SLIGHTLY_EARLY] = [2] * floor((max_depth + 1) / 2) + [1] * ceil((max_depth + 1) / 2)
        self.weighting_function[itemBias.NOBIAS] = [1] * (max_depth + 1)
        self.weighting_function[itemBias.SLIGHTLY_LATE] = [1] * floor((max_depth + 1) / 2) + [2] * ceil((max_depth + 1) / 2)
        self.weighting_function[itemBias.LATE]  = weight_list_maker(0)
        self.weighting_function[itemBias.VERY_LATE]  = weight_list_maker(1)
        self.weighting_function[itemBias.SUPER_LATE]  = weight_list_maker(2)
        self.weighting_function[itemBias.NIGHTMARE]  = weight_list_maker(3)

    
    def get_rarity_weightings(self):
        return self.weighting_function

class WeightDistributions:

    def __init__(self, max_depth: int):
        self.weighting_function: dict[itemDifficulty, dict[itemRarity, list[int]]] = {
            itemDifficulty.NORMAL: {}
        }
        for rarity in itemRarity:
            self.weighting_function[itemDifficulty.NORMAL][itemRarity(rarity)] = [1] * (max_depth + 1)

        self.weighting_function[itemDifficulty.SLIGHTLY_HARD] = {
            itemRarity.COMMON: [1] * (max_depth + 1),
            itemRarity.UNCOMMON: [1] * (max_depth + 1),
            itemRarity.RARE: [1] * floor((max_depth + 1) / 2) + [2] * ceil((max_depth + 1) / 2),
            itemRarity.MYTHIC: [1] * floor((max_depth + 1) / 2) + [2] * ceil((max_depth + 1) / 2)
        }
        self.weighting_function[itemDifficulty.SLIGHTLY_EASY] = {
            itemRarity.COMMON: [1] * (max_depth + 1),
            itemRarity.UNCOMMON: [1] * (max_depth + 1),
            itemRarity.RARE: [2] * floor((max_depth + 1) / 2) + [1] * ceil((max_depth + 1) / 2),
            itemRarity.MYTHIC: [2] * floor((max_depth + 1) / 2) + [1] * ceil((max_depth + 1) / 2)
        }

        easier_difficulties = [
            itemDifficulty.SUPEREASY,
            itemDifficulty.EASY
        ]
        for diff_index, difficulty in enumerate(easier_difficulties):
            diff_index_modded = (1 - diff_index) - 0.25
            self.weighting_function[difficulty] = {
                itemRarity.COMMON: [
                    max(1, floor(10.0 / pow(x / (max_depth / 3.0), diff_index_modded * .5 + 1)))
                    for x in range(max_depth + 1, 0, -1)
                ],
                itemRarity.UNCOMMON: [
                    max(1, floor(3.0 / pow(x / (max_depth / 3.0), diff_index_modded * .5 + 1)))
                    for x in range(max_depth + 1, 0, -1)
                ],
                itemRarity.RARE: [
                    max(1, floor(3.0 / pow((max_depth + 2 - x) / (max_depth / 3.0), diff_index_modded * .5 + 1)))
                    for x in range(max_depth + 1, 0, -1)],
                itemRarity.MYTHIC: [
                    max(1, floor(10.0 / pow((max_depth + 2 - x) / (max_depth / 3.0), diff_index_modded * .5 + 1)))
                    for x in range(max_depth + 1, 0, -1)
                ]
            }

        harder_difficulties = [
            itemDifficulty.HARD,
            itemDifficulty.VERYHARD,
            itemDifficulty.INSANE,
            itemDifficulty.NIGHTMARE
        ]
        for diff_index, difficulty in enumerate(harder_difficulties):
            diff_index_modded = diff_index - 0.25
            self.weighting_function[difficulty] = {
                itemRarity.COMMON: [
                    max(1, floor(10.0 / pow(x / (max_depth / 3.0), diff_index_modded * .5 + 1)))
                    for x in range(1, max_depth + 2)
                ],
                itemRarity.UNCOMMON: [
                    max(1, floor(3.0 / pow(x / (max_depth / 3.0), diff_index_modded * .5 + 1)))
                    for x in range(1, max_depth + 2)
                ],
                itemRarity.RARE: [
                    max(1, floor(3.0 / pow((max_depth + 2 - x) / (max_depth / 3.0), diff_index_modded * .5 + 1)))
                    for x in range(1, max_depth + 2)],
                itemRarity.MYTHIC: [
                    max(1, floor(10.0 / pow((max_depth + 2 - x) / (max_depth / 3.0), diff_index_modded * .5 + 1)))
                    for x in range(1, max_depth + 2)
                ]
            }

        # for diff_index,difficulty in enumerate(["Super Easy","Easy","Hard","Very Hard","Insane","Nightmare"]):
        #     print(self.weighting_function[difficulty][itemRarity.COMMON])
        #     print(self.weighting_function[difficulty][itemRarity.UNCOMMON])
        #     print(self.weighting_function[difficulty][itemRarity.RARE])
        #     print(self.weighting_function[difficulty][itemRarity.MYTHIC])
        #     print("--------------------")

    def get_rarity_weighting(self, difficulty: itemDifficulty):
        if difficulty not in self.weighting_function:
            raise ValueError(f"Unknown item placement difficulty {difficulty}")
        return self.weighting_function[difficulty]


class LocationWeights:

    def __init__(self, settings: RandomizerSettings, locations: Locations, reverse_locations: Locations):
        self.regular_rando = settings.regular_rando
        self.reverse_rando = settings.reverse_rando
        self.split_levels = settings.split_levels
        self.level_offsets = DreamWeaponOffsets()
        self.max_level = settings.max_level_checks
        self.level_depths: dict[int, int] = {}

        # Regular
        # -------------------------------
        self.location_depths: dict[KH2Location, int] = {}
        hops = locations.location_graph.get_hops(START_NODE)
        location_type_maxes: dict[locationType, int] = {}
        max_hops = 17  # hard coding max hops to unify depths between reverse and regular rando
        # for hop in hops:
        #     max_hops = max(max_hops,hop[1])
        for node, distance in hops:
            for loc in locations.locations_for_node(node):
                primary_type = loc.LocationTypes[0]
                if primary_type not in location_type_maxes:
                    location_type_maxes[primary_type] = distance
                else:
                    location_type_maxes[primary_type] = max(distance, location_type_maxes[primary_type])

        for node, distance in hops:
            for loc in locations.locations_for_node(node):
                primary_type = loc.LocationTypes[0]
                if location_type_maxes[primary_type] != 0:
                    scaled_depth = floor((distance * 1.0 / location_type_maxes[primary_type]) * max_hops)
                else:
                    scaled_depth = distance
                self.location_depths[loc] = scaled_depth
                if loc.LocationCategory is locationCategory.LEVEL:
                    self.level_depths[loc.LocationId] = scaled_depth

        # Reverse
        # -------------------------------
        self.reverse_location_depths: dict[KH2Location, int] = {}
        hops = reverse_locations.location_graph.get_hops(START_NODE)
        reverse_location_type_maxes: dict[locationType, int] = {}
        # max_hops = 0
        # for hop in hops:
        #     max_hops = max(max_hops,hop[1])
        for node, distance in hops:
            for loc in reverse_locations.locations_for_node(node):
                primary_type = loc.LocationTypes[0]
                if primary_type not in reverse_location_type_maxes:
                    reverse_location_type_maxes[primary_type] = distance
                else:
                    reverse_location_type_maxes[primary_type] = max(distance, reverse_location_type_maxes[primary_type])

        for node, distance in hops:
            for loc in reverse_locations.locations_for_node(node):
                primary_type = loc.LocationTypes[0]
                if reverse_location_type_maxes[primary_type] != 0:
                    scaled_depth = floor((distance * 1.0 / reverse_location_type_maxes[primary_type]) * max_hops)
                else:
                    scaled_depth = distance
                self.reverse_location_depths[loc] = scaled_depth

        self.base_weights = SimplifiedWeightDistributions(max_hops).get_rarity_weightings()
        self.weights = {}
        # default to normal weighting
        for item_type in itemType:
            self.weights[item_type] = self.base_weights[itemBias.NOBIAS]

        self.weights[itemType.FORM] = self.base_weights[settings.form_weights]
        self.weights[itemType.STORYUNLOCK] = self.base_weights[settings.unlock_weights]

        self.weights[itemType.FIRE] = self.base_weights[settings.magic_weights]
        self.weights[itemType.BLIZZARD] = self.base_weights[settings.magic_weights]
        self.weights[itemType.THUNDER] = self.base_weights[settings.magic_weights]
        self.weights[itemType.CURE] = self.base_weights[settings.magic_weights]
        self.weights[itemType.MAGNET] = self.base_weights[settings.magic_weights]
        self.weights[itemType.REFLECT] = self.base_weights[settings.magic_weights]

        self.weights[itemType.TORN_PAGE] = self.base_weights[settings.page_weights]

        self.weights[itemType.SUMMON] = self.base_weights[settings.summon_weights]

        self.weights[itemType.REPORT] = self.base_weights[settings.report_weights]

        self.weights[itemType.PROOF_OF_PEACE] = self.base_weights[settings.proof_weights]
        self.weights[itemType.PROOF_OF_CONNECTION] = self.base_weights[settings.proof_weights]
        self.weights[itemType.PROOF_OF_NONEXISTENCE] = self.base_weights[settings.proof_weights]

        self.weights[itemType.PROMISE_CHARM] = self.base_weights[settings.promise_charm_weights]

    def get_depth(self, location: KH2Location) -> Union[int, tuple[int, int]]:
        """
        Returns either a single depth (for levels or regular or reverse rando) or a tuple of
        (regular depth, reverse depth) if both regular and reverse rando.
        """
        if location.LocationCategory is locationCategory.LEVEL:
            loc_id = location.LocationId
            if self.split_levels:
                sword_depth = self.level_depths[loc_id]
                shield_depth = self.level_depths[self.level_offsets.get_shield_level(self.max_level, loc_id)]
                staff_depth = self.level_depths[self.level_offsets.get_staff_level(self.max_level, loc_id)]
                return (sword_depth + shield_depth + staff_depth) // 3
            else:
                return self.level_depths[loc_id]
        else:
            if self.regular_rando and self.reverse_rando:
                return self.location_depths[location], self.reverse_location_depths[location]
            elif self.regular_rando:
                return self.location_depths[location]
            else:
                return self.reverse_location_depths[location]

    def get_weight(self, rarity: itemRarity, location: KH2Location) -> int:
        raise AssertionError("Deprecated weighting function")
        """ Returns the weight that should be used for an item of the given rarity at the given location. """
        depth_or_depths = self.get_depth(location)
        rarity_weights = self.weights[rarity]
        if isinstance(depth_or_depths, int):
            return rarity_weights[depth_or_depths]
        elif isinstance(depth_or_depths, tuple):
            regular_weight = rarity_weights[depth_or_depths[0]]
            reverse_weight = rarity_weights[depth_or_depths[1]]
            return (regular_weight + reverse_weight) // 2
        

    def get_weight(self, item_type: itemType, location: KH2Location) -> int:
        """ Returns the weight that should be used for an item of the given type at the given location. """
        depth_or_depths = self.get_depth(location)
        rarity_weights = self.weights[item_type]
        if isinstance(depth_or_depths, int):
            return rarity_weights[depth_or_depths]
        elif isinstance(depth_or_depths, tuple):
            regular_weight = rarity_weights[depth_or_depths[0]]
            reverse_weight = rarity_weights[depth_or_depths[1]]
            return (regular_weight + reverse_weight) // 2



ABILITY_POOL = "ability_pool"
ACCESSIBILITY = "seed_accessibility"
ACCESSORIES_IN_ITEM_POOL = "accessories_in_item_pool"
ALLOW_PROOF_HINTING = "allow_proof_hinting"
ALLOW_REPORT_HINTING = "allow_report_hinting"
ANTIFORM = "antiform"
ARMOR_IN_ITEM_POOL = "armor_in_item_pool"
AS_DATA_SPLIT = "as_data_reward_split"
ATLANTICA_TUTORIAL_SKIP = "skip_atlantica_tutorial"
AUTO_EQUIP_START_ABILITIES = "auto_equip"
BATTLE_LEVEL_OFFSET = "battle_level_offset"
BATTLE_LEVEL_RANDO = "battle_level_rando"
BATTLE_LEVEL_RANDOM_MAX = "battle_level_random_max"
BATTLE_LEVEL_RANDOM_MIN = "battle_level_random_min"
BATTLE_LEVEL_RANGE = "battle_level_range"
BLOCK_COR_SKIP = "block_cor_skip"
BLOCK_SHAN_YU_SKIP = "block_shan_yu_skip"
CHAIN_LOGIC = "chain_logic"
CHAIN_LOGIC_LENGTH = "chain_logic_length"
MAX_CHAIN_LOGIC_LENGTH = "max_chain_logic_length"
CHAIN_LOGIC_MIN_TERRA = "chain_logic_min_terra"
CHAIN_LOGIC_TERRA = "chain_logic_terra"
CHESTS_MATCH_ITEM = "chests_match_item"
COMMAND_MENU = "command_menu"
COSTUME_RANDO = "costume_rando"
CRITICAL_BONUS_REWARDS = "critical_bonus_rewards"
CUPS_GIVE_XP = "cups_give_xp"
DISABLE_FINAL_FORM = "disable_final_form_behavior"
DONALD_AP = "donald_ap"
DOUBLE_STAT_GROWTH_MIN_LEVEL = "double_stat_growth_min_level"
DOUBLE_STAT_GROWTH_MAX_LEVEL = "double_stat_growth_max_level"
EMBLEM_NUM_REQUIRED = "emblem_num_required"
EMBLEM_NUM_AVAILABLE = "emblem_num_available"
ENABLE_PROMISE_CHARM = "enable_promise_charm"
ENDPIC_RANDO = "endpic_rando"
FAST_URNS = "fast_urns"
FIFTY_AP_BOOSTS = "fifty_ap_boosts"
FINAL_DOOR_REQUIREMENT = "final_door_requirement"
FINAL_EXP_CURVE = "final_exp_curve"
FINAL_EXP_MULTIPLIER = "final_exp_multiplier"
GARDEN_OF_ASSEMBLAGE_REWARDS = "garden_of_assemblage_rewards"
GLASS_CANNON = "glass_cannon"
GLOBAL_JACKPOT = "global_jackpot"
GLOBAL_LUCKY = "global_lucky"
GOOFY_AP = "goofy_ap"
HINT_SYSTEM = "hint_system"
HINTABLE_CHECKS = "hintable_checks"
ITEM_PLACEMENT_DIFFICULTY = "item_placement_difficulty"
ITEMPIC_RANDO = "itempic_rando"
JUNK_ITEMS = "junk_items"
KEYBLADE_ACTION_ABILITIES = "keyblade_action_abilities"
KEYBLADE_SUPPORT_ABILITIES = "keyblade_support_abilities"
KEYBLADE_MAX_STAT = "keyblade_max_stat"
KEYBLADE_MIN_STAT = "keyblade_min_stat"
KEYBLADE_RANDO = "keyblade_rando"
KEYBLADE_RANDO_ALLOW_DUPLICATES = "keyblade_rando_allow_duplicates"
KEYBLADE_RANDO_INCLUDE_EFFECTS = "keyblade_rando_include_effects"
KEYBLADES_LOCK_CHESTS = "keyblades_lock_chests"
LIMIT_EXP_CURVE = "limit_exp_curve"
LIMIT_EXP_MULTIPLIER = "limit_exp_multiplier"
MAPS_IN_ITEM_POOL = "maps_in_item_pool"
MASTER_EXP_CURVE = "master_exp_curve"
MASTER_EXP_MULTIPLIER = "master_exp_multiplier"
MISC_LOCATIONS_WITH_REWARDS = "misc_locations_with_rewards"
MUNNY_IN_ITEM_POOL = "munny_in_item_pool"
MUSIC_RANDO_ENABLED_PC = "music_rando_enabled_pc"
MUSIC_RANDO_PC_ALLOW_DUPLICATES = "music_rando_pc_allow_duplicates"
MUSIC_RANDO_PC_DMCA_SAFE = "music_rando_pc_dmca_safe"
MUSIC_RANDO_PC_INCLUDE_BBS = "music_rando_pc_include_all_bbs"
MUSIC_RANDO_PC_INCLUDE_CUSTOM = "music_rando_pc_include_custom"
MUSIC_RANDO_PC_INCLUDE_DDD = "music_rando_pc_include_all_ddd"
MUSIC_RANDO_PC_INCLUDE_KH1 = "music_rando_pc_include_all_kh1"
MUSIC_RANDO_PC_INCLUDE_KH2 = "music_rando_pc_include_all_kh2"
MUSIC_RANDO_PC_INCLUDE_RECOM = "music_rando_pc_include_all_recom"
MUSIC_RANDO_PC_USE_CATEGORIES = "music_rando_pc_use_categories"
NIGHTMARE_LOGIC = "nightmare_logic"
OBJECTIVE_RANDO_NUM_REQUIRED = "objective_rando_num_required"
OBJECTIVE_RANDO_NUM_AVAILABLE = "objective_rando_num_available"
OBJECTIVE_RANDO_POOL = "objective_rando_pool"
PARTY_MEMBER_RANDO = "party_member_rando"
POINTS_ABILITY = "points_abilities"
POINTS_ABILITY_COLLECT = "points_ability_collect"
POINTS_AUX = "points_aux"
POINTS_BONUS = "points_bonus"
POINTS_BOSS_AS = "points_boss_as"
POINTS_BOSS_DATA = "points_boss_data"
POINTS_BOSS_FINAL = "points_boss_final"
POINTS_BOSS_NORMAL = "points_boss_normal"
POINTS_BOSS_SEPHIROTH = "points_boss_sephiroth"
POINTS_BOSS_TERRA = "points_boss_terra"
POINTS_COMPLETE = "points_complete"
POINTS_DEATH = "points_death"
POINTS_FORM = "points_forms"
POINTS_FORM_COLLECT = "points_form_collect"
POINTS_FORMLV = "points_formlv"
POINTS_KEYBLADES = "points_keyblades"
POINTS_MAGIC = "points_magic"
POINTS_MAGIC_COLLECT = "points_magic_collect"
POINTS_PAGE = "points_page"
POINTS_PAGE_COLLECT = "points_page_collect"
POINTS_POUCHES_COLLECT = "points_pouches_collect"
POINTS_PROOF = "points_proofs"
POINTS_PROOF_COLLECT = "points_proof_collect"
POINTS_REPORT = "points_report"
POINTS_REPORT_COLLECT = "points_report_collect"
POINTS_SUMMON = "points_summon"
POINTS_SUMMON_COLLECT = "points_summon_collect"
POINTS_VISIT = "points_visit"
POINTS_VISIT_COLLECT = "points_visit_collect"
PR_MAP_SKIP = "pr_map_skip"
PREVENT_SELF_HINTING = "prevent_self_hinting"
PROGRESSION_HINTS = "progression_hints"
PROGRESSION_HINTS_COMPLETE_BONUS = "progression_complete_bonus"
PROGRESSION_HINTS_REPORT_BONUS = "progression_report_bonus"
PROGRESSION_HINTS_REVEAL_END = "progression_points_reveal_end"
PROGRESSION_POINT_SELECT = "progression_point_select"
PROMISE_CHARM_DEPTH = "promise_charm_depth"
PROOF_DEPTH = "proof_depth"
RANDO_THEMED_TEXTURES = "rando_themed_textures"
RANDOMIZED_SETTINGS = "randomized_settings"
RANDOMIZED_SETTINGS_ENABLED = "randomized_settings_enabled"
RECOLOR_TEXTURES = "recolor_textures"
RECOLOR_TEXTURES_COMPRESS = "recolor_textures_compress"
RECOLOR_TEXTURES_INCLUDE_EXTRAS = "recolor_textures_include_extras"
RECOLOR_TEXTURES_KEEP_CACHE = "recolor_textures_keep_cache"
RECIPES_IN_ITEM_POOL = "recipes_in_item_pool"
REMOVE_CUTSCENES = "remove_cutscenes"
REMOVE_DAMAGE_CAP = "remove_damage_cap"
REMOVE_POPUPS = "remove_popups"
REMOVE_WARDROBE_ANIMATION = "remove_wardrobe_animation"
REPORT_DEPTH = "report_depth"
REPORTS_REVEAL = "reports_reveal"
JOURNAL_HINTS_ABILITIES = "journal_hints_abilities"
RETRY_DARK_THORN = "retry_dark_thorn"
RETRY_DFX = "retry_data_final_xemnas"
REVEAL_COMPLETE = "reveal_complete"
REVENGE_LIMIT_RANDO = "revenge_limit_rando"
RICH_ENEMIES = "rich_enemies"
ROOM_TRANSITION_IMAGES = "room_transition_images"
ROXAS_ABILITIES_ENABLED = "roxas_abilities_enabled"
SCORE_MODE = "score_mode"
SHOP_BOOSTS = "shop_boosts"
SHOP_ELIXIRS = "shop_elixirs"
SHOP_KEYBLADES = "shop_keyblades"
SHOP_RECOVERIES = "shop_recoveries"
SHOP_REPORTS = "shop_reports"
SHOP_UNLOCKS = "shop_unlocks"
SKIP_CARPET_ESCAPE = "skip_carpet_escape"
SOFTLOCK_CHECKING = "softlock_checking"
SORA_AP = "sora_ap"
SORA_EXP_CURVE = "sora_exp_curve"
SORA_EXP_MULTIPLIER = "sora_exp_multiplier"
SORA_LEVELS = "sora_levels"
SORA_STR_RATE = "sora_strength_rate"
SORA_DEF_RATE = "sora_defense_rate"
SORA_MAG_RATE = "sora_magic_rate"
SORA_AP_RATE = "sora_ap_rate"
SPLIT_LEVELS = "split_levels"
SPOILER_REVEAL_TYPES = "spoiler_reveal_types"
STARTING_INVENTORY = "starting_inventory"
STARTING_MOVEMENT = "starting_movement"
STARTING_REPORTS = "starting_reports"
STARTING_MAGIC_FIRE = "starting_magic_fire"
STARTING_MAGIC_BLIZZARD = "starting_magic_blizzard"
STARTING_MAGIC_THUNDER = "starting_magic_thunder"
STARTING_MAGIC_CURE = "starting_magic_cure"
STARTING_MAGIC_MAGNET = "starting_magic_magnet"
STARTING_MAGIC_REFLECT = "starting_magic_reflect"
STARTING_KEYBLADES = "starting_keyblades"
STARTING_ABILITIES = "starting_abilities"
STARTING_PAGES = "starting_pages"
STARTING_DRIVES = "starting_drives"
STARTING_ITEMS = "starting_items"
STARTING_UNLOCKS_AG = "starting_unlocks_ag"
STARTING_UNLOCKS_BC = "starting_unlocks_bc"
STARTING_UNLOCKS_DC = "starting_unlocks_dc"
STARTING_UNLOCKS_HB = "starting_unlocks_hb"
STARTING_UNLOCKS_HT = "starting_unlocks_ht"
STARTING_UNLOCKS_LOD = "starting_unlocks_lod"
STARTING_UNLOCKS_OC = "starting_unlocks_oc"
STARTING_UNLOCKS_PL = "starting_unlocks_pl"
STARTING_UNLOCKS_PR = "starting_unlocks_pr"
STARTING_UNLOCKS_SP = "starting_unlocks_sp"
STARTING_UNLOCKS_STT = "starting_unlocks_stt"
STARTING_UNLOCKS_TT = "starting_unlocks_tt"
STARTING_UNLOCKS_TWTNW = "starting_unlocks_twtnw"
STARTING_VISIT_MODE = "starting_visit_mode"
STARTING_VISIT_RANDOM_MAX = "starting_visit_random_max"
STARTING_VISIT_RANDOM_MIN = "starting_visit_random_min"
STATSANITY = "statsanity"
STORY_UNLOCK_CATEGORY = "story_unlock_category"
STORY_UNLOCK_DEPTH = "story_unlock_depth"
SUMMON_EXP_CURVE = "summon_exp_curve"
SUMMON_EXP_MULTIPLIER = "summon_exp_multiplier"
SUPERBOSSES_WITH_REWARDS = "superbosses_with_rewards"
TEXTURE_RECOLOR_SETTINGS = "texture_recolor_settings"
UNLIMITED_MP = "unlimited_mp"
VALOR_EXP_CURVE = "valor_exp_curve"
VALOR_EXP_MULTIPLIER = "valor_exp_multiplier"
WEIGHTED_FORMS = "weighted_forms"
WEIGHTED_MAGIC = "weighted_magic"
WEIGHTED_PAGES = "weighted_pages"
WEIGHTED_PROMISE_CHARM = "weighted_promise_charm"
WEIGHTED_PROOFS = "weighted_proofs"
WEIGHTED_REPORTS = "weighted_reports"
WEIGHTED_SUMMONS = "weighted_summons"
WEIGHTED_UNLOCKS = "weighted_unlocks"
WISDOM_EXP_CURVE = "wisdom_exp_curve"
WISDOM_EXP_MULTIPLIER = "wisdom_exp_multiplier"
WORLDS_WITH_REWARDS = "worlds_with_rewards"
YEET_THE_BEAR = "yeet_the_bear"
DONALD_MELEE_ATTACKS_KNOCKBACK_TYPE = "donald_melee_attacks_knockback_type"
DONALD_FIRE_KNOCKBACK_TYPE = "donald_fire_knockback_type"
DONALD_BLIZZARD_KNOCKBACK_TYPE = "donald_blizzard_knockback_type"
DONALD_THUNDER_KNOCKBACK_TYPE = "donald_thunder_knockback_type"
GOOFY_MELEE_ATTACKS_KNOCKBACK_TYPE = "goofy_melee_attacks_knockback_type"
GOOFY_BASH_KNOCKBACK_TYPE = "goofy_bash_knockback_type"
GOOFY_TURBO_KNOCKBACK_TYPE = "goofy_turbo_knockback_type"
GOOFY_TORNADO_KNOCKBACK_TYPE = "goofy_tornado_knockback_type"
COMPANION_DAMAGE_TOGGLE = "companion_damage_toggle"
COMPANION_KILL_BOSS = "companion_kill_boss"
JACK_SKELLINGTON_MELEE_ATTACKS_KNOCKBACK_TYPE = "jack_skellington_melee_attacks_knockback_type"
SIMBA_MELEE_ATTACKS_KNOCKBACK_TYPE = "simba_melee_attacks_knockback_type"
ALADDIN_MELEE_ATTACKS_KNOCKBACK_TYPE = "aladdin_melee_attacks_knockback_type"
PING_MULAN_MELEE_ATTACKS_KNOCKBACK_TYPE = "ping_mulan_melee_attacks_knockback_type"
BEAST_MELEE_ATTACKS_KNOCKBACK_TYPE = "beast_melee_attacks_knockback_type"
TRON_MELEE_ATTACKS_KNOCKBACK_TYPE = "tron_melee_attacks_knockback_type"
JACK_SPARROW_MELEE_ATTACKS_KNOCKBACK_TYPE = "jack_sparrow_melee_attacks_knockback_type"
RIKU_MELEE_ATTACKS_KNOCKBACK_TYPE = "riku_melee_attacks_knockback_type"
AURON_MELEE_ATTACKS_KNOCKBACK_TYPE = "auron_melee_attacks_knockback_type"
RIKU_DARK_AURA_KNOCKBACK_TYPE = "riku_dark_aura_knockback_type"
RIKU_DARK_SHIELD_KNOCKBACK_TYPE = "riku_dark_shield_knockback_type"
AURON_DIVIDER_KNOCKBACK_TYPE = "auron_divider_knockback_type"
JACK_SPARROW_NO_MERCY_KNOCKBACK_TYPE = "jack_sparrow_no_mercy_knockback_type"
JACK_SPARROW_RAIN_STORM_KNOCKBACK_TYPE = "jack_sparrow_rain_time_knockback_type"
JACK_SPARROW_BONE_SMASH_KNOCKBACK_TYPE = "jack_sparrow_bone_smash_knockback_type"
TRON_SCOUTING_DISK_KNOCKBACK_TYPE = "tron_scouting_disk_knockback_type"
TRON_PULSING_THUNDER_KNOCKBACK_TYPE = "tron_pulsing_thunder_knockback_type"
JACK_SKELLINGTON_BLAZING_FURY_KNOCKBACK_TYPE = "jack_skellington_blazing_fury_knockback_type"
JACK_SKELLINGTON_ICY_TERROR_KNOCKBACK_TYPE = "jack_skellington_icy_terror_knockback_type"
JACK_SKELLINGTON_BOLTS_OF_SORROW_KNOCKBACK_TYPE = "jack_skellington_bolts_of_sorrow_knockback_type"
SIMBA_FIERCE_CLAW_KNOCKBACK_TYPE = "simba_fierce_claw_knockback_type"
SIMBA_GROUNDSHAKER_KNOCKBACK_TYPE = "simba_groundshaker_knockback_type"
ALADDIN_SLASH_FRENZY_KNOCKBACK_TYPE = "aladdin_slash_frenzy_knockback_type"
ALADDIN_QUICKPLAY_KNOCKBACK_TYPE = "aladdin_quickplay_knockback_type"
MULAN_MUSHU_FIRE_KNOCKBACK_TYPE = "mulan_mushu_fire_knockback_type"
MULAN_FLAMETONGUE_KNOCKBACK_TYPE = "mulan_flametongue_knockback_type"
BEAST_SHOUT_KNOCKBACK_TYPE = "beast_shout_knockback_type"
BEAST_RUSH_KNOCKBACK_TYPE = "beast_rush_knockback_type"



from List.configDict import locationType
from Module import encoding


class ProgressionPoints():
    def __init__(self):
        self.gen_to_tracker_location_map = {
            locationType.Level:"Levels",
            locationType.FormLevel:"Drives",
            locationType.STT:"SimulatedTwilightTown",
            locationType.TT:"TwilightTown",
            locationType.HB:"HollowBastion",
            locationType.BC:"BeastsCastle",
            locationType.OC:"OlympusColiseum",
            locationType.Agrabah:"Agrabah",
            locationType.LoD:"LandofDragons",
            locationType.HUNDREDAW:"HundredAcreWood",
            locationType.PL:"PrideLands",
            locationType.Atlantica:"Atlantica",
            locationType.DC:"DisneyCastle",
            locationType.HT:"HalloweenTown",
            locationType.PR:"PortRoyal",
            locationType.SP:"SpaceParanoids",
            locationType.TWTNW:"TWTNW",
            locationType.CoR:"CavernofRemembrance"
        }
        self.initialize_points()
        self.initialize_point_thresholds()

    def get_world_options(self):
        return list(self.gen_to_tracker_location_map.keys())
    def get_num_cp_for_world(self,world:locationType):
        return len(self.points[world])
    def get_cp_for_world(self,world:locationType,index:int):
        return self.points[world][index]
    def get_cp_label_for_world(self,world:locationType,index:int):
        human_readable_labels = {
            "Levels":["Level 10", "Level 20", "Level 30", "Level 40", "Level 50"],
            "Drives":["Drive Level 2", "Drive Level 3", "Drive Level 4", "Drive Level 5", "Drive Level 6", "Drive Level 7"],
            "SimulatedTwilightTown":[ "STT Enter", "Olette Munny Pouch", "Twilight Thorn", "Axel 1", "Setzer", "Mansion Computer Room", "Axel 2", "Data Roxas" ],
            "TwilightTown":[ "Twilight Town Enter", "Station Dusk Fight", "Mysterious Tower", "Sandlot Fight", "Mansion Fight", "Betwixt and Between", "Data Axel" ],
            "HollowBastion":[ "Hollow Bastion Enter", "Bailey Fight", "Ansem's Study Computer", "Corridors Fight", "Dancers Fight", "Demyx", "Final Fantasy Fights", "1000 Heartless", "Sephiroth", "Data Demyx" ],
            "LandofDragons":[ "Land of Dragons Enter", "Mission 3", "Mountain Trail", "Cave Fight", "Summit Fight", "Shan Yu", "Antechamber Nobodies", "Stormrider", "Data Xigbar" ],
            "BeastsCastle":[ "Beast's Castle Enter", "Thresholder", "Beast's Room", "Dark Thorn", "Ballroom Dragoons Fight", "Xaldin", "Data Xaldin" ],
            "OlympusColiseum":[ "Olympus Coliseum Enter", "Cerberus", "Phil's Urns", "Demyx", "Pete", "Hydra", "Auron Statue", "Hades", "AS Zexion", "Data Zexion" ],
            "DisneyCastle":[ "Disney Castle Enter", "Minnie Escort", "Old Pete", "Timeless River Windows", "Boat Pete", "Pete", "AS Marluxia", "Data Marluxia", "Lingering Will" ],
            "PortRoyal":[ "Port Royal Enter", "Town Fight", "1 Minute Fight", "Boat Medallion", "Boat Barrels", "Barbossa", "Grim Reaper 1", "First Medallion Gambler", "Grim Reaper 2", "Data Luxord" ],
            "Agrabah":[ "Agrabah Enter", "Abu Minigame", "Chasm of Challenges", "Treasure Room", "Twin Lords", "Carpet Magic", "Genie Jafar", "AS Lexaeus", "Data Lexaeus" ],
            "HalloweenTown":[ "Halloween Town Enter", "Candy Cane Lane Fight", "Prison Keeper", "Oogie Boogie", "Lock, Shock, Barrel", "Presents Minigame", "Experiment", "AS Vexen", "Data Vexen" ],
            "PrideLands":[ "Pride Lands Enter", "Talking to Simba", "Hyenas 1", "Scar", "Hyenas 2", "Groundshaker", "Data Saix" ],
            "SpaceParanoids":[ "Space Paranoids Enter", "Screens", "Hostile Program", "Solar Sailer Fight", "MCP", "AS Larxene", "Data Larxene" ],
            "HundredAcreWood":[ "Pooh's Howse Enter", "Piglet's Howse Complete", "Rabbit's Howse Complete", "Kanga's Howse Complete", "Spooky Cave Complete", "Starry Hill Complete" ],
            "Atlantica":[ "Atlantica Enter", "Ursula's Revenge", "A New Day is Dawning" ],
            "TWTNW":[ "TWTNW Enter", "Roxas", "Xigbar", "Luxord", "Saix", "Xemnas 1", "Data Xemnas" ],
            "CavernofRemembrance":[ "CoR Enter", "First Fight", "Steam Valves", "Second Fight", "Transport to Remembrance" ]
        }
        return human_readable_labels[self.gen_to_tracker_location_map[world]][index]
    def set_cp_for_world(self,world:locationType,index:int, value:int):
        self.points[world][index] = value

    def initialize_point_thresholds(self):
        self.point_thresholds = [ 3, 3, 3, 4, 4, 4, 5, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8] # max 18 (17 worlds + creations)

    def initialize_points(self):
        self.points = {
            locationType.Level:[ 0, 0, 1, 1, 2 ],
            locationType.FormLevel:[ 0, 0, 0, 1, 1, 1 ],
            locationType.STT:[ 0, 1, 2, 1, 1, 1, 1, 6 ],
            locationType.TT:[ 0, 1, 1, 2, 2, 2, 6 ],
            locationType.HB:[ 0, 1, 0, 2, 1, 2, 2, 2, 5, 6 ],
            locationType.BC:[ 0, 2, 1, 2, 1, 3, 6 ],
            locationType.OC:[ 0, 2, 0, 2, 0, 2, 2, 3, 3, 6 ],
            locationType.Agrabah:[ 0, 1, 1, 1, 2, 1, 3, 4, 6 ],
            locationType.LoD:[ 0, 1, 1, 1, 1, 2, 0, 3, 6 ],
            locationType.HUNDREDAW:[ 0, 1, 1, 1, 2, 4 ],
            locationType.PL:[ 0, 1, 1, 2, 1, 3, 6 ],
            locationType.Atlantica:[  0, 5, 3 ],
            locationType.DC:[ 0, 1, 1, 2, 1, 2, 5, 6, 7 ],
            locationType.HT:[ 0, 1, 2, 2, 1, 1, 3, 3, 6 ],
            locationType.PR:[ 0, 1, 1, 0, 1, 2, 1, 1, 3, 6 ],
            locationType.SP:[ 0, 1, 2, 2, 2, 4, 6 ],
            locationType.TWTNW:[ 0, 2, 3, 2, 2, 2, 7 ],
            locationType.CoR:[0, 4, 0, 4, 7]
        }

    def get_compressed(self) -> str:
        keys = list(self.points.keys())

        threshold_characters = ''
        for p in self.point_thresholds:
            threshold_characters += str(p)

        points_characters = ''
        for k in keys:
            for i in range(len(self.points[k])):
                points_characters += str(self.points[k][i])

        threshold_result = encoding.v2r(int(threshold_characters))
        points_result = encoding.v2r(int(points_characters))
        result = threshold_result + '+' + points_result

        return result

    def set_uncompressed(self, compressed_string: str):
        keys = list(self.points.keys())

        split_strings = compressed_string.split('+')

        threshold_count = len(self.point_thresholds)
        decoded_threshold = str(encoding.r2v(split_strings[0]))
        threshold_characters = self._prepend_zeroes(decoded_threshold, expected_length=threshold_count)
        for i in range(threshold_count):
            self.set_point_threshold(i, int(threshold_characters[i]))

        points_expected_length = 0
        for k in keys:
            points_expected_length = points_expected_length + len(self.points[k])
        decoded_points = str(encoding.r2v(split_strings[1]))
        points_characters = self._prepend_zeroes(decoded_points, expected_length=points_expected_length)

        current_index = 0
        for k in keys:
            location_points: list[int] = self.points[k]
            for i in range(len(location_points)):
                location_points[i] = int(points_characters[current_index])
                current_index += 1

        return True

    def get_point_threshold(self,index):
        return self.point_thresholds[index]
    
    def set_point_threshold(self,index,value):
        self.point_thresholds[index] = value

    def get_hint_thresholds(self,world_count):
        output = [self.point_thresholds[x] for x in range(len(self.point_thresholds)) if x < world_count]
        return output

    def get_points_json(self):
        output = {}
        for l in locationType:
            if l in self.points:
                output[self.gen_to_tracker_location_map[l]] = self.points[l]
        return output

    def settings_spoiler_json(self) -> dict[str, list[int]]:
        output = {"Hint Thresholds": self.point_thresholds}
        for location in locationType:
            if location in self.points:
                output[locationType(location).value] = self.points[location]
        return output

    @staticmethod
    def _prepend_zeroes(value: str, expected_length: int) -> str:
        difference = expected_length - len(value)
        return ('0' * difference) + value



import json
from pathlib import Path
from typing import Optional

import yaml

AUTOSAVE_FOLDER = "auto-save"
PRESET_FOLDER = "presets"


def read_app_config() -> dict:
    config_path = Path('randomizer-config.json').absolute()
    if config_path.is_file():
        with open(config_path, encoding='utf-8') as config_file:
            return json.load(config_file)
    else:
        return {}

def read_boss_enemy_override_files() -> str:
    result_string = ""
    location_override_path = Path('override_enemies.yaml').absolute()
    if location_override_path.is_file():
        with open(location_override_path, encoding='utf-8') as location_override:
            result_string += location_override.read()
    enemy_override_path = Path('override_enemies.yaml').absolute()
    if enemy_override_path.is_file():
        with open(enemy_override_path, encoding='utf-8') as enemy_override:
            result_string += enemy_override.read()
    return result_string
            

def write_app_config(config: dict):
    config_path = Path('randomizer-config.json').absolute()
    with open(config_path, mode='w', encoding='utf-8') as config_file:
        json.dump(config, config_file, indent=4)


def update_app_config(key: str, value):
    randomizer_config = read_app_config()
    randomizer_config[key] = value
    write_app_config(randomizer_config)


def remove_app_config(key: str):
    randomizer_config = read_app_config()
    if key in randomizer_config:
        del randomizer_config[key]
        write_app_config(randomizer_config)


def auto_save_folder() -> Path:
    return Path(AUTOSAVE_FOLDER)


def read_openkh_path() -> Optional[Path]:
    randomizer_config = read_app_config()
    openkh_folder = randomizer_config.get("openkh_folder", "to-nowhere")
    if openkh_folder is None:
        return None

    openkh_path = Path(openkh_folder)
    if openkh_path.is_dir():
        return openkh_path
    else:
        return None


def write_openkh_path(selected_directory):
    update_app_config('openkh_folder', selected_directory)


def read_custom_music_path() -> Optional[Path]:
    randomizer_config = read_app_config()
    custom_music_folder = randomizer_config.get("custom_music_folder", "to-nowhere")
    if custom_music_folder is None:
        return None

    custom_music_path = Path(custom_music_folder)
    if custom_music_path.is_dir():
        return custom_music_path
    else:
        return None


def write_custom_music_path(selected_directory):
    update_app_config('custom_music_folder', selected_directory)


def read_custom_visuals_path() -> Optional[Path]:
    randomizer_config = read_app_config()
    custom_visuals_folder = randomizer_config.get("custom_visuals_folder", "to-nowhere")
    if custom_visuals_folder is None:
        return None

    custom_visuals_path = Path(custom_visuals_folder)
    if custom_visuals_path.is_dir():
        return custom_visuals_path
    else:
        return None


def write_custom_visuals_path(selected_directory):
    update_app_config('custom_visuals_folder', selected_directory)


def extracted_data_path() -> Optional[Path]:
    """Returns the path to extracted game data."""
    openkh_path = read_openkh_path()
    if openkh_path is None:
        return None

    mods_manager_yml_path = openkh_path / "mods-manager.yml"
    if not mods_manager_yml_path.is_file():
        return None

    with open(mods_manager_yml_path, encoding="utf-8") as mod_manager_file:
        mod_manager_yaml = yaml.safe_load(mod_manager_file)
        game_data_path = mod_manager_yaml.get("gameDataPath", "to-nowhere")
        if game_data_path is None:
            # Maybe something weird happened in the Mods Manager setup? Should have a gameDataPath.
            # Regardless, nothing we can do here.
            return None

        extracted_data = Path(game_data_path)
        if extracted_data.is_dir():
            return extracted_data
        else:
            # Has a gameDataPath configured, but it's not being detected as a directory
            return None



import random
from itertools import chain

from Class import settingkey
from Class.exceptions import SettingsException
from Class.seedSettings import SeedSettings, makeKHBRSettings
from List import experienceValues, ObjectiveList
from List.configDict import (
    DisableFinalOption,
    ObjectivePoolOption,
    locationType,
    itemType,
    locationDepth,
    StartingMovementOption,
    SoraLevelOption,
    ItemAccessibilityOption,
    itemRarity,
    SoftlockPreventionOption,
    AbilityPoolOption,
    expCurve,
    StartingVisitMode,
    FinalDoorRequirement,
)
from List.hashTextEntries import generate_hash_icons
from List.inventory import report, proof, form, storyunlock, magic, misc
from List.inventory.keyblade import get_locking_keyblade_names
from Module.modifier import SeedModifier
from Module.progressionPoints import ProgressionPoints


class RandomizerSettings:
    excludeFrom50 = list(
        chain(
            [
                1,
                3,
                5,
                6,
                8,
                11,
                13,
                16,
                18,
                19,
                21,
                22,
                24,
                26,
                27,
                29,
                31,
                33,
                35,
                37,
                38,
                40,
                42,
                43,
                45,
                47,
                49,
            ],
            range(51, 100),
        )
    )
    excludeFrom99 = [
        1,
        2,
        3,
        4,
        5,
        6,
        8,
        10,
        11,
        13,
        14,
        16,
        18,
        19,
        21,
        22,
        24,
        26,
        27,
        29,
        30,
        32,
        34,
        35,
        37,
        38,
        40,
        42,
        43,
        45,
        46,
        48,
        50,
        51,
        52,
        54,
        55,
        56,
        57,
        58,
        60,
        61,
        62,
        63,
        64,
        66,
        67,
        68,
        69,
        70,
        71,
        72,
        74,
        75,
        76,
        77,
        78,
        79,
        80,
        81,
        82,
        83,
        84,
        86,
        87,
        88,
        89,
        90,
        91,
        92,
        93,
        94,
        95,
        96,
        97,
        98,
    ]

    def __init__(
        self,
        seed_name: str,
        spoiler_log: bool,
        ui_version: str,
        ui_settings: SeedSettings,
        seed_string: str,
        boss_enemy_overrides: str = None, 
    ):
        self.ui_settings = ui_settings
        self.settings_spoiler_json = ui_settings.settings_spoiler_json()
        self.seed_string = seed_string
        self.crit_mode = ui_settings.get(settingkey.CRITICAL_BONUS_REWARDS)
        self.item_accessibility = ItemAccessibilityOption(
            ui_settings.get(settingkey.ACCESSIBILITY)
        )

        include_list = []
        vanilla_list = []
        include_list_keys = [
            (settingkey.CRITICAL_BONUS_REWARDS, "Critical Bonuses"),
            (settingkey.GARDEN_OF_ASSEMBLAGE_REWARDS, "Garden of Assemblage"),
        ]
        for key in include_list_keys:
            if ui_settings.get(key[0]):
                include_list.append(key[1])
        worlds_with_rewards = ui_settings.get(settingkey.WORLDS_WITH_REWARDS)
        vanilla_worlds = []
        if isinstance(worlds_with_rewards[0], list):
            vanilla_worlds = worlds_with_rewards[1]
            worlds_with_rewards = worlds_with_rewards[0]
        for location in worlds_with_rewards:
            include_list.append(locationType[location].value)
        for location in vanilla_worlds:
            vanilla_list.append(locationType[location].value)
        for location in ui_settings.get(settingkey.SUPERBOSSES_WITH_REWARDS):
            include_list.append(locationType[location].value)
        for location in ui_settings.get(settingkey.MISC_LOCATIONS_WITH_REWARDS):
            include_list.append(locationType[location].value)
        self.enabledLocations: list[locationType] = [
            l for l in locationType if l in include_list
        ]
        self.vanillaLocations: list[locationType] = [
            l for l in locationType if l in vanilla_list
        ]
        self.disabledLocations: list[locationType] = [
            l
            for l in locationType
            if l not in include_list
            and l not in [locationType.Mush13, locationType.WeaponSlot, locationType.Creations] # added creations here since we have all the subtypes covered
        ]

        level_setting = SoraLevelOption(ui_settings.get(settingkey.SORA_LEVELS))
        self.vanilla_level_one = False
        if level_setting == SoraLevelOption.LEVEL_1:
            self.max_level_checks = 1
            if locationType.Level in self.enabledLocations:
                raise SettingsException(
                    "Please choose between Junk or Vanilla Checks when doing Level 1"
                )
            elif locationType.Level in self.vanillaLocations:
                self.vanilla_level_one = True
        elif level_setting == SoraLevelOption.LEVEL_50:
            self.max_level_checks = 50
        elif level_setting == SoraLevelOption.LEVEL_99:
            self.max_level_checks = 99
        else:
            raise SettingsException("Invalid Level choice")
        
        self.min_double_stat_level = ui_settings.get(settingkey.DOUBLE_STAT_GROWTH_MIN_LEVEL)
        self.max_double_stat_level = ui_settings.get(settingkey.DOUBLE_STAT_GROWTH_MAX_LEVEL)

        self.split_levels: bool = ui_settings.get(settingkey.SPLIT_LEVELS)
        self.battle_level_rando: str = ui_settings.get(settingkey.BATTLE_LEVEL_RANDO)
        self.battle_level_offset: int = ui_settings.get(settingkey.BATTLE_LEVEL_OFFSET)
        self.battle_level_range: int = ui_settings.get(settingkey.BATTLE_LEVEL_RANGE)
        self.battle_level_random_min_max: tuple[int, int] = (
            ui_settings.get(settingkey.BATTLE_LEVEL_RANDOM_MIN),
            ui_settings.get(settingkey.BATTLE_LEVEL_RANDOM_MAX)
        )

        self.starting_inventory_ids: list[int] = []
        self.starting_inventory_ids.extend([
            int(value) for value in ui_settings.get(settingkey.STARTING_KEYBLADES)])
        self.starting_inventory_ids.extend([
            int(value) for value in ui_settings.get(settingkey.STARTING_DRIVES)])
        self.starting_inventory_ids.extend([
            int(value) for value in ui_settings.get(settingkey.STARTING_ITEMS)])
        self.starting_inventory_ids.extend([
            int(value) for value in ui_settings.get(settingkey.STARTING_ABILITIES)])
        self.starting_inventory_ids.extend([magic.Fire.id for _ in range(ui_settings.get(settingkey.STARTING_MAGIC_FIRE))])
        self.starting_inventory_ids.extend([magic.Blizzard.id for _ in range(ui_settings.get(settingkey.STARTING_MAGIC_BLIZZARD))])
        self.starting_inventory_ids.extend([magic.Thunder.id for _ in range(ui_settings.get(settingkey.STARTING_MAGIC_THUNDER))])
        self.starting_inventory_ids.extend([magic.Cure.id for _ in range(ui_settings.get(settingkey.STARTING_MAGIC_CURE))])
        self.starting_inventory_ids.extend([magic.Magnet.id for _ in range(ui_settings.get(settingkey.STARTING_MAGIC_MAGNET))])
        self.starting_inventory_ids.extend([magic.Reflect.id for _ in range(ui_settings.get(settingkey.STARTING_MAGIC_REFLECT))])
        self.starting_inventory_ids.extend([misc.TornPages.id for _ in range(ui_settings.get(settingkey.STARTING_PAGES))])

        self.starting_growth_option = StartingMovementOption(
            ui_settings.get(settingkey.STARTING_MOVEMENT)
        )
        self.starting_report_count: int = ui_settings.get(settingkey.STARTING_REPORTS)

        self.starting_visit_mode: StartingVisitMode = StartingVisitMode[ui_settings.get(settingkey.STARTING_VISIT_MODE)]
        self.starting_visit_random_range: tuple[int, int] = (
            ui_settings.get(settingkey.STARTING_VISIT_RANDOM_MIN),
            ui_settings.get(settingkey.STARTING_VISIT_RANDOM_MAX)
        )
        self.starting_unlocks_per_world: dict[locationType, int] = {
            locationType.Agrabah: ui_settings.get(settingkey.STARTING_UNLOCKS_AG),
            locationType.BC: ui_settings.get(settingkey.STARTING_UNLOCKS_BC),
            locationType.DC: ui_settings.get(settingkey.STARTING_UNLOCKS_DC),
            locationType.HB: ui_settings.get(settingkey.STARTING_UNLOCKS_HB),
            locationType.HT: ui_settings.get(settingkey.STARTING_UNLOCKS_HT),
            locationType.LoD: ui_settings.get(settingkey.STARTING_UNLOCKS_LOD),
            locationType.OC: ui_settings.get(settingkey.STARTING_UNLOCKS_OC),
            locationType.PL: ui_settings.get(settingkey.STARTING_UNLOCKS_PL),
            locationType.PR: ui_settings.get(settingkey.STARTING_UNLOCKS_PR),
            locationType.SP: ui_settings.get(settingkey.STARTING_UNLOCKS_SP),
            locationType.STT: ui_settings.get(settingkey.STARTING_UNLOCKS_STT),
            locationType.TT: ui_settings.get(settingkey.STARTING_UNLOCKS_TT),
            locationType.TWTNW: ui_settings.get(settingkey.STARTING_UNLOCKS_TWTNW),
        }

        self.form_weights = ui_settings.get(settingkey.WEIGHTED_FORMS)
        self.unlock_weights = ui_settings.get(settingkey.WEIGHTED_UNLOCKS)
        self.magic_weights = ui_settings.get(settingkey.WEIGHTED_MAGIC)
        self.page_weights = ui_settings.get(settingkey.WEIGHTED_PAGES)
        self.summon_weights = ui_settings.get(settingkey.WEIGHTED_SUMMONS)
        self.report_weights = ui_settings.get(settingkey.WEIGHTED_REPORTS)
        self.proof_weights = ui_settings.get(settingkey.WEIGHTED_PROOFS)
        self.promise_charm_weights = ui_settings.get(settingkey.WEIGHTED_PROMISE_CHARM)

        self.extended_placement_logic: bool = ui_settings.get(
            settingkey.NIGHTMARE_LOGIC
        )
        self.story_unlock_rarity: itemRarity = ui_settings.get(
            settingkey.STORY_UNLOCK_CATEGORY
        )

        softlock_option = SoftlockPreventionOption(
            ui_settings.get(settingkey.SOFTLOCK_CHECKING)
        )
        self.regular_rando: bool = softlock_option in [
            SoftlockPreventionOption.DEFAULT,
            SoftlockPreventionOption.BOTH,
        ]
        self.reverse_rando: bool = softlock_option in [
            SoftlockPreventionOption.REVERSE,
            SoftlockPreventionOption.BOTH,
        ]

        self.level_stat_pool = SeedModifier.level_up_stat_pool_weighted(
            str_rate=ui_settings.get(settingkey.SORA_STR_RATE),
            mag_rate=ui_settings.get(settingkey.SORA_MAG_RATE),
            def_rate=ui_settings.get(settingkey.SORA_DEF_RATE),
            ap_rate=ui_settings.get(settingkey.SORA_AP_RATE),
        )

        self.junk_pool: list[int] = [
            int(item_id) for item_id in ui_settings.get(settingkey.JUNK_ITEMS)
        ]
        if len(self.junk_pool) == 0:
            raise SettingsException("Need at least one junk item in the pool")

        self.sora_ap: int = ui_settings.get(settingkey.SORA_AP)
        self.donald_ap: int = ui_settings.get(settingkey.DONALD_AP)
        self.goofy_ap: int = ui_settings.get(settingkey.GOOFY_AP)
        self.ability_pool_option = AbilityPoolOption(
            ui_settings.get(settingkey.ABILITY_POOL)
        )
        self.promiseCharm: bool = ui_settings.get(settingkey.ENABLE_PROMISE_CHARM)
        self.auto_equip_abilities: bool = ui_settings.get(
            settingkey.AUTO_EQUIP_START_ABILITIES
        )
        self.tt1_jailbreak: bool = True  # ui_settings.get(settingkey.TT1_JAILBREAK)
        self.pureblood: bool = True  # ui_settings.get(settingkey.PUREBLOOD)
        self.randoRandoEnabled = ui_settings.get(settingkey.RANDOMIZED_SETTINGS_ENABLED)
        self.antiform: bool = ui_settings.get(settingkey.ANTIFORM)
        self.fifty_ap: bool = ui_settings.get(settingkey.FIFTY_AP_BOOSTS)
        self.hintsType: str = ui_settings.get(settingkey.HINT_SYSTEM)
        self.reportDepth = locationDepth(ui_settings.get(settingkey.REPORT_DEPTH))
        self.proofDepth = locationDepth(ui_settings.get(settingkey.PROOF_DEPTH))
        self.promiseCharmDepth = locationDepth(ui_settings.get(settingkey.PROMISE_CHARM_DEPTH))
        self.storyDepth = locationDepth(ui_settings.get(settingkey.STORY_UNLOCK_DEPTH))

        if softlock_option == SoftlockPreventionOption.BOTH:
            if self.proofDepth in [
                locationDepth.FirstBoss,
                locationDepth.LastStoryBoss,
                locationDepth.FirstVisit,
            ]:
                raise SettingsException(
                    f"Setting proof depth to {self.proofDepth} will contradict either regular or reverse rando. Please use another setting"
                )
            if self.reportDepth in [
                locationDepth.FirstBoss,
                locationDepth.LastStoryBoss,
                locationDepth.FirstVisit,
            ]:
                raise SettingsException(
                    f"Setting report depth to {self.reportDepth} will contradict either regular or reverse rando. Please use another setting"
                )
            if self.storyDepth in [
                locationDepth.FirstBoss,
                locationDepth.LastStoryBoss,
                locationDepth.FirstVisit,
            ]:
                raise SettingsException(
                    f"Setting visit unlock depth to {self.storyDepth} will contradict either regular or reverse rando. Please use another setting"
                )

        self.prevent_self_hinting: bool = ui_settings.get(
            settingkey.PREVENT_SELF_HINTING
        )
        self.allow_proof_hinting: bool = ui_settings.get(settingkey.ALLOW_PROOF_HINTING)
        self.allow_report_hinting: bool = ui_settings.get(
            settingkey.ALLOW_REPORT_HINTING
        )
        self.keyblade_support_abilities: list[int] = [
            int(ability_id)
            for ability_id in ui_settings.get(settingkey.KEYBLADE_SUPPORT_ABILITIES)
        ]
        self.keyblade_action_abilities: list[int] = [
            int(ability_id)
            for ability_id in ui_settings.get(settingkey.KEYBLADE_ACTION_ABILITIES)
        ]
        self.keyblade_min_stat: int = ui_settings.get(settingkey.KEYBLADE_MIN_STAT)
        self.keyblade_max_stat: int = ui_settings.get(settingkey.KEYBLADE_MAX_STAT)

        self.sora_exp_multiplier: float = ui_settings.get(
            settingkey.SORA_EXP_MULTIPLIER
        )
        self.valor_exp_multiplier: float = ui_settings.get(
            settingkey.VALOR_EXP_MULTIPLIER
        )
        self.wisdom_exp_multiplier: float = ui_settings.get(
            settingkey.WISDOM_EXP_MULTIPLIER
        )
        self.limit_exp_multiplier: float = ui_settings.get(
            settingkey.LIMIT_EXP_MULTIPLIER
        )
        self.master_exp_multiplier: float = ui_settings.get(
            settingkey.MASTER_EXP_MULTIPLIER
        )
        self.final_exp_multiplier: float = ui_settings.get(
            settingkey.FINAL_EXP_MULTIPLIER
        )
        self.summon_exp_multiplier: float = ui_settings.get(
            settingkey.SUMMON_EXP_MULTIPLIER
        )
        self.sora_exp_curve: expCurve = expCurve.from_name(
            ui_settings.get(settingkey.SORA_EXP_CURVE)
        )
        self.valor_exp_curve: expCurve = expCurve.from_name(
            ui_settings.get(settingkey.VALOR_EXP_CURVE)
        )
        self.wisdom_exp_curve: expCurve = expCurve.from_name(
            ui_settings.get(settingkey.WISDOM_EXP_CURVE)
        )
        self.limit_exp_curve: expCurve = expCurve.from_name(
            ui_settings.get(settingkey.LIMIT_EXP_CURVE)
        )
        self.master_exp_curve: expCurve = expCurve.from_name(
            ui_settings.get(settingkey.MASTER_EXP_CURVE)
        )
        self.final_exp_curve: expCurve = expCurve.from_name(
            ui_settings.get(settingkey.FINAL_EXP_CURVE)
        )
        self.summon_exp_curve: expCurve = expCurve.from_name(
            ui_settings.get(settingkey.SUMMON_EXP_CURVE)
        )

        self.as_data_split: bool = ui_settings.get(settingkey.AS_DATA_SPLIT)
        self.chests_match_item: bool = ui_settings.get(settingkey.CHESTS_MATCH_ITEM)
        self.skip_carpet_escape: bool = ui_settings.get(settingkey.SKIP_CARPET_ESCAPE)

        if self.reverse_rando and not self.as_data_split:
            raise SettingsException(
                "Can't run reverse rando without the as/data split code"
            )

        self.enemy_options = makeKHBRSettings(seed_name, ui_settings)

        self.random_seed: str = seed_name
        self.spoiler_log: bool = spoiler_log
        self.ui_version: str = ui_version
        self.boss_enemy_overrides: str = boss_enemy_overrides if boss_enemy_overrides and (self.enemy_options["boss"] != "Disabled" or self.enemy_options["enemy"] != "Disabled") else ""
        self.create_full_seed_string()
        self.seedHashIcons: list[str] = generate_hash_icons()

        self.statSanity: bool = ui_settings.get(settingkey.STATSANITY)
        self.yeetTheBear: bool = ui_settings.get(settingkey.YEET_THE_BEAR)
        self.chainLogic: bool = ui_settings.get(settingkey.CHAIN_LOGIC)
        self.chainLogicMinLength: int = ui_settings.get(settingkey.CHAIN_LOGIC_LENGTH)
        self.chainLogicMaxLength: int = ui_settings.get(settingkey.MAX_CHAIN_LOGIC_LENGTH)

        if (
            self.proofDepth in [locationDepth.FirstVisit, locationDepth.FirstBoss]
            and self.chainLogic
        ):
            raise SettingsException(
                "Chain logic is not compatible with first visit proofs"
            )
        if (
            self.storyDepth not in [locationDepth.Anywhere, locationDepth.NonSuperboss]
            and self.chainLogic
        ):
            raise SettingsException(
                "Chain logic is only compatible with visit unlock depth non-data and anywhere"
            )
        if self.chainLogic and self.regular_rando and self.reverse_rando:
            raise SettingsException(
                "Can't do chain logic with both regular and reverse rando"
            )

        self.keyblades_unlock_chests: bool = ui_settings.get(settingkey.KEYBLADES_LOCK_CHESTS)

        self.roxas_abilities_enabled: bool = ui_settings.get(
            settingkey.ROXAS_ABILITIES_ENABLED
        )
        self.disable_antiform: bool = ui_settings.get(settingkey.DISABLE_FINAL_FORM) == DisableFinalOption.NO_ANTIFORM
        self.disable_final_form: bool = ui_settings.get(settingkey.DISABLE_FINAL_FORM) == DisableFinalOption.NO_FINAL
        self.block_cor_skip: bool = ui_settings.get(settingkey.BLOCK_COR_SKIP)
        self.block_shan_yu_skip: bool = ui_settings.get(settingkey.BLOCK_SHAN_YU_SKIP)
        self.pr_map_skip: bool = ui_settings.get(settingkey.PR_MAP_SKIP)
        self.atlantica_skip: bool = ui_settings.get(settingkey.ATLANTICA_TUTORIAL_SKIP)
        self.wardrobe_skip: bool = ui_settings.get(settingkey.REMOVE_WARDROBE_ANIMATION)
        self.include_maps: bool = ui_settings.get(settingkey.MAPS_IN_ITEM_POOL)
        self.include_recipes: bool = ui_settings.get(settingkey.RECIPES_IN_ITEM_POOL)
        self.include_munny_pouches: bool = ui_settings.get(settingkey.MUNNY_IN_ITEM_POOL)
        self.include_accessories: bool = ui_settings.get(settingkey.ACCESSORIES_IN_ITEM_POOL)
        self.include_armor: bool = ui_settings.get(settingkey.ARMOR_IN_ITEM_POOL)
        self.remove_popups: bool = ui_settings.get(settingkey.REMOVE_POPUPS)

        self.global_jackpot: int = ui_settings.get(settingkey.GLOBAL_JACKPOT)
        self.global_lucky: int = ui_settings.get(settingkey.GLOBAL_LUCKY)
        self.rich_enemies: bool = ui_settings.get(settingkey.RICH_ENEMIES)
        self.unlimited_mp: bool = ui_settings.get(settingkey.UNLIMITED_MP)
        self.fast_urns: bool = ui_settings.get(settingkey.FAST_URNS)

        self.shop_elixirs: bool = ui_settings.get(settingkey.SHOP_ELIXIRS)
        self.shop_recoveries: bool = ui_settings.get(settingkey.SHOP_RECOVERIES)
        self.shop_boosts: bool = ui_settings.get(settingkey.SHOP_BOOSTS)
        self.shop_keyblades: bool = ui_settings.get(settingkey.SHOP_KEYBLADES)
        self.shop_unlocks: int = ui_settings.get(settingkey.SHOP_UNLOCKS)
        self.shop_reports: int = ui_settings.get(settingkey.SHOP_REPORTS)

        self.point_hint_values: dict[str, int] = {
            "proof": ui_settings.get(settingkey.POINTS_PROOF),
            "form": ui_settings.get(settingkey.POINTS_FORM),
            "magic": ui_settings.get(settingkey.POINTS_MAGIC),
            "summon": ui_settings.get(settingkey.POINTS_SUMMON),
            "ability": ui_settings.get(settingkey.POINTS_ABILITY),
            "keyblade": ui_settings.get(settingkey.POINTS_KEYBLADES),
            "page": ui_settings.get(settingkey.POINTS_PAGE),
            "report": ui_settings.get(settingkey.POINTS_REPORT),
            "visit": ui_settings.get(settingkey.POINTS_VISIT),
            "bonus": ui_settings.get(settingkey.POINTS_BONUS),
            "complete": ui_settings.get(settingkey.POINTS_COMPLETE),
            "formlv": ui_settings.get(settingkey.POINTS_FORMLV),
            "other": ui_settings.get(settingkey.POINTS_AUX),
            "boss_as": ui_settings.get(settingkey.POINTS_BOSS_AS),
            "boss_datas": ui_settings.get(settingkey.POINTS_BOSS_DATA),
            "boss_sephi": ui_settings.get(settingkey.POINTS_BOSS_SEPHIROTH),
            "boss_terra": ui_settings.get(settingkey.POINTS_BOSS_TERRA),
            "boss_final": ui_settings.get(settingkey.POINTS_BOSS_FINAL),
            "boss_other": ui_settings.get(settingkey.POINTS_BOSS_NORMAL),
            "deaths": ui_settings.get(settingkey.POINTS_DEATH),
            "collection_magic": ui_settings.get(settingkey.POINTS_MAGIC_COLLECT),
            "collection_page": ui_settings.get(settingkey.POINTS_PAGE_COLLECT),
            "collection_pouches": ui_settings.get(settingkey.POINTS_POUCHES_COLLECT),
            "collection_proof": ui_settings.get(settingkey.POINTS_PROOF_COLLECT),
            "collection_form": ui_settings.get(settingkey.POINTS_FORM_COLLECT),
            "collection_summon": ui_settings.get(settingkey.POINTS_SUMMON_COLLECT),
            "collection_ability": ui_settings.get(settingkey.POINTS_ABILITY_COLLECT),
            "collection_report": ui_settings.get(settingkey.POINTS_REPORT_COLLECT),
            "collection_visit": ui_settings.get(settingkey.POINTS_VISIT_COLLECT),
        }

        self.progression_hints: bool = ui_settings.get(settingkey.PROGRESSION_HINTS)
        self.progression_world_complete_bonus: int = ui_settings.get(
            settingkey.PROGRESSION_HINTS_COMPLETE_BONUS
        )
        self.progression_report_bonus: int = ui_settings.get(
            settingkey.PROGRESSION_HINTS_REPORT_BONUS
        )
        self.progression_reveal_all: bool = ui_settings.get(
            settingkey.PROGRESSION_HINTS_REVEAL_END
        )

        self.shop_hintable = (
            (self.shop_unlocks > 0)
            or (self.shop_reports > 0)
            or self.shop_keyblades
            or locationType.Puzzle in include_list
            or locationType.SYNTH in include_list
        )
        if (self.shop_unlocks > 0) or (self.shop_reports > 0) or self.shop_keyblades:
            include_list.append(locationType.SYNTH)

        prog_points = ProgressionPoints()
        prog_points.set_uncompressed(
            ui_settings.get(settingkey.PROGRESSION_POINT_SELECT)
        )
        self.progression_hint_settings = prog_points.get_points_json()
        num_worlds = (
            len(vanilla_worlds)
            + len(worlds_with_rewards)
            + (1 if self.shop_hintable else 0)
        )
        self.progression_hint_settings["HintCosts"] = prog_points.get_hint_thresholds(
            num_worlds
        )
        self.progression_hint_settings["WorldCompleteBonus"] = [
            self.progression_world_complete_bonus
        ]
        self.progression_hint_settings["ReportBonus"] = [self.progression_report_bonus]
        self.progression_hint_settings["FinalXemnasReveal"] = [
            1 if self.progression_reveal_all else 0
        ]

        self.revealComplete: bool = ui_settings.get(settingkey.REVEAL_COMPLETE)
        self.revealMode: str = ui_settings.get(settingkey.REPORTS_REVEAL)
        self.journal_hints: str = ui_settings.get(settingkey.JOURNAL_HINTS_ABILITIES)

        final_door_requirement = FinalDoorRequirement[ui_settings.get(settingkey.FINAL_DOOR_REQUIREMENT)]
        self.objective_rando: bool = final_door_requirement is FinalDoorRequirement.OBJECTIVES
        self.emblems: bool = final_door_requirement is FinalDoorRequirement.EMBLEMS
        self.num_objectives_needed = ui_settings.get(settingkey.OBJECTIVE_RANDO_NUM_REQUIRED)
        self.max_objectives_available = ui_settings.get(settingkey.OBJECTIVE_RANDO_NUM_AVAILABLE)
        self.num_emblems_needed = ui_settings.get(settingkey.EMBLEM_NUM_REQUIRED)
        self.max_emblems_available = ui_settings.get(settingkey.EMBLEM_NUM_AVAILABLE)
        self.available_objectives = ObjectiveList.get_full_objective_list()

        self.objective_pool_type = ui_settings.get(settingkey.OBJECTIVE_RANDO_POOL)
        if self.objective_pool_type == ObjectivePoolOption.BOSSES.name:
            self.available_objectives = [o for o in self.available_objectives if o.Type == ObjectiveList.ObjectiveType.BOSS]
        elif self.objective_pool_type == ObjectivePoolOption.NOBOSSES.name:
            self.available_objectives = [o for o in self.available_objectives if o.Type == ObjectiveList.ObjectiveType.WORLDPROGRESS or o.Type == ObjectiveList.ObjectiveType.FIGHT]
        elif self.objective_pool_type == ObjectivePoolOption.HITLIST.name:
            self.available_objectives = [o for o in self.available_objectives if o.Difficulty==ObjectiveList.ObjectiveDifficulty.LATEST ]
        elif self.objective_pool_type == ObjectivePoolOption.LASTSTORY.name:
            self.available_objectives = ObjectiveList.get_last_story_objectives()

        if self.max_objectives_available > len(self.available_objectives):
            raise SettingsException("Not enough enabled objectives for the requested available pool. Turn on more objectives or lower available")

        self.hintable_check_types: list[str] = [
            item_type for item_type in ui_settings.get(settingkey.HINTABLE_CHECKS)
        ]
        self.spoiler_hint_values: list[str] = [
            item_type for item_type in ui_settings.get(settingkey.SPOILER_REVEAL_TYPES)
        ]
        if self.revealComplete:
            self.spoiler_hint_values.append("complete")
        if self.revealMode != "Disabled":
            self.spoiler_hint_values.append(self.revealMode)
            if (
                self.revealMode == "bossreports"
                and ui_settings.get("boss") == "Disabled"
            ):
                raise SettingsException(
                    "Can't use report hint bosses option without boss randomization."
                )
        if (
            self.hintsType == "Spoiler"
            and self.revealMode == "Disabled"
            and self.progression_hints
        ):
            raise SettingsException(
                "Can't use progression hints with full spoiler hints, set the report reveal mode to reports please"
            )

        self.hiscore_mode: bool = ui_settings.get(settingkey.SCORE_MODE)

        self.tracker_includes: list[str] = []
        
        if self.objective_rando:
            self.tracker_includes.append("objectives")
        if self.progression_hints:
            self.tracker_includes.append("ProgressionHints")
        if self.vanilla_level_one or locationType.Level in self.disabledLocations:
            self.tracker_includes.append("Level1Mode")
        self.tracker_includes.append(level_setting)
        if self.roxas_abilities_enabled:
            self.tracker_includes.append("better_stt")
        if self.as_data_split:
            self.tracker_includes.append("Data Split")
        if self.starting_report_count == 13:
            self.tracker_includes.append("library")
        if self.hiscore_mode:
            self.tracker_includes.append("ScoreMode")
        if self.split_levels:
            self.tracker_includes.append("Dream Weapon Matters")

        if (
            self.hintsType in ["JSmartee", "Path"]
            and "proof" not in self.hintable_check_types
        ):
            raise SettingsException(
                "Jsmartee and Path hints really need proofs hintable."
            )
        self.important_checks = []
        self.spoiler_reveal_checks = []
        self.populate_check_type_list(self.hintable_check_types, self.important_checks)
        self.populate_check_type_list(
            self.spoiler_hint_values, self.spoiler_reveal_checks
        )

        for check_type in [
            "magic",
            "proof",
            "form",
            "page",
            "summon",
            "visit",
            "ability",
            "other",
            "report",
            "keyblade",
        ]:
            if check_type not in self.hintable_check_types:
                self.point_hint_values[check_type] = 0

        # making tracker includes use all worlds and
        for l in locationType:
            if l.value in self.enabledLocations:
                if l.value != "Level":  # don't duplicate the level info
                    self.tracker_includes.append(l.value)
            elif l.value in self.vanillaLocations:
                if l.value != "Level":  # don't duplicate the level info
                    self.tracker_includes.append(l.value)
            elif self.randoRandoEnabled:
                if l.value != "Level":  # don't duplicate the level info
                    self.tracker_includes.append(l.value)

        # Add synth if it's not already there and if there's anything in the shop.
        # Otherwise, the tracker may not have a location for the shop items to get tracked to.
        if self.shop_hintable and locationType.SYNTH not in self.tracker_includes:
            self.tracker_includes.append(locationType.SYNTH)

        # putting this in the settings object to allow us to turn it off as a safety valve
        self.dummy_forms = True

        self.validateSettings()

    def populate_check_type_list(self, hintable_checks_list, check_list):
        if "magic" in hintable_checks_list:
            check_list += [
                itemType.FIRE,
                itemType.BLIZZARD,
                itemType.THUNDER,
                itemType.CURE,
                itemType.REFLECT,
                itemType.MAGNET,
            ]
        if "proof" in hintable_checks_list:
            check_list += [
                itemType.PROOF_OF_NONEXISTENCE,
                itemType.PROOF_OF_CONNECTION,
                itemType.PROOF_OF_PEACE,
                itemType.PROMISE_CHARM,
            ]
        if "form" in hintable_checks_list:
            check_list += [itemType.FORM, "Anti-Form"]
        if "page" in hintable_checks_list:
            check_list += [itemType.TORN_PAGE]
        if "report" in hintable_checks_list:
            check_list += [itemType.REPORT]
        if "summon" in hintable_checks_list:
            check_list += [itemType.SUMMON]
        if "visit" in hintable_checks_list:
            check_list += [itemType.STORYUNLOCK]
        if "keyblade" in hintable_checks_list:
            check_list += get_locking_keyblade_names()
        if "ability" in hintable_checks_list:
            check_list += ["Second Chance", "Once More"]
        if "other" in hintable_checks_list:
            check_list += [
                itemType.TROPHY,
                itemType.MANUFACTORYUNLOCK,
                itemType.OCSTONE,
                itemType.MUNNY_POUCH,
            ]

    def create_full_seed_string(self):
        seed_string_from_all_inputs = (
            self.random_seed
            + str(self.spoiler_log)
            + self.ui_version
            + str(self.ui_settings.settings_string()) 
            + self.boss_enemy_overrides
        )
        self.full_rando_seed = seed_string_from_all_inputs
        random.seed(seed_string_from_all_inputs)

    def validateSettings(self):
        boss_depths = [
            locationDepth.FirstBoss,
            locationDepth.LastStoryBoss,
            locationDepth.Superbosses,
        ]
        if self.reportDepth == self.proofDepth and self.reportDepth in boss_depths:
            raise SettingsException(
                "Proof depth and report depth can't be set to the same boss category"
            )
        if self.storyDepth == self.proofDepth and self.proofDepth in boss_depths:
            raise SettingsException(
                "Proof depth and visit unlock depth can't be set to the same boss category"
            )
        if self.reportDepth == self.storyDepth and self.reportDepth in boss_depths:
            raise SettingsException(
                "Visit unlock depth and report depth can't be set to the same boss category"
            )

        if locationType.TTR in self.enabledLocations and not self.statSanity:
            raise SettingsException(
                "Enabling Transport to Remembrance when not in Statsanity is incorrect. Enable Statsanity or disable TTR."
            )

        if self.chainLogic:
            if len(self.vanillaLocations) > 0:
                raise SettingsException(
                    "Currently can't do chain logic and vanilla worlds. Sorry about that. "
                )

            if proof.ProofOfNonexistence.id in self.starting_inventory_ids:
                raise SettingsException(
                    "Cannot use chain logic if starting with Proof of Nonexistence"
                )
            
            if self.objective_rando:
                raise SettingsException(
                    "Can't use chain logic and objective rando at the same time"
                )
            if self.emblems:
                raise SettingsException(
                    "Can't use chain logic and emblems at the same time"
                )

        if (
            self.ability_pool_option != AbilityPoolOption.DEFAULT
            and len(self.vanillaLocations) > 0
        ):
            pass
            # raise SettingsException("Currently can't do randomized ability pools and vanilla worlds. Sorry about that. ")

        if "report" not in self.hintable_check_types:
            if (
                self.hintsType in ["JSmartee", "Points", "Spoiler", "Path"]
                and not self.progression_hints
                and self.revealMode != "Disabled"
            ):
                raise SettingsException(f"The selected hint system {'Progression' if self.progression_hints else ''} {self.hintsType}  needs to have reports trackable")

        max_reports = len(report.all_reports())
        if self.starting_report_count + self.shop_reports > max_reports:
            raise SettingsException(
                f"Starting Ansem Reports + Ansem Reports in shop is more than {max_reports}"
            )
        
        if self.objective_rando and self.max_objectives_available < self.num_objectives_needed:
            raise SettingsException(
                f"Can't set required objectives {self.num_objectives_needed} less than objectives in the pool {self.max_objectives_available}"
            )
        if self.emblems and self.max_emblems_available < self.num_emblems_needed:
            raise SettingsException(
                f"Can't set required emblems {self.num_emblems_needed} less than emblems in the pool {self.max_emblems_available}"
            )


        starting_visit_mode = self.starting_visit_mode
        starting_visit_items_count = 0
        if starting_visit_mode is StartingVisitMode.FIRST:
            starting_visit_items_count = len(storyunlock.all_story_unlocks())
        elif starting_visit_mode is StartingVisitMode.ALL:
            starting_visit_items_count = len(storyunlock.all_individual_story_unlocks())
        elif starting_visit_mode is StartingVisitMode.RANDOM:
            # Going to be pessimistic here and assume worst-case of max random
            starting_visit_items_count = self.starting_visit_random_range[1]
        elif starting_visit_mode is StartingVisitMode.SPECIFIC:
            for _, count in self.starting_unlocks_per_world.items():
                starting_visit_items_count = starting_visit_items_count + count
        elif starting_visit_mode is StartingVisitMode.CUSTOM:
            for _, count in self.starting_unlocks_per_world.items():
                starting_visit_items_count = starting_visit_items_count + count
            starting_visit_items_count = starting_visit_items_count + self.starting_visit_random_range[1]

        max_unlocks = len(storyunlock.all_individual_story_unlocks())
        if starting_visit_items_count > max_unlocks:
            raise SettingsException(
                f"Starting Visit Unlocks plus Maximum Visits Available is more than {max_unlocks}"
            )

        if starting_visit_items_count + self.shop_unlocks > max_unlocks:
            raise SettingsException(
                f"Starting Visit Unlocks plus Visit Unlocks in shop is more than {max_unlocks}"
            )

        if self.yeetTheBear:
            if proof.ProofOfNonexistence.id in self.starting_inventory_ids:
                raise SettingsException(
                    "Cannot require Yeet the Bear if starting with Proof of Nonexistence"
                )

            if locationType.HUNDREDAW not in self.enabledLocations:
                raise SettingsException(
                    "Cannot require Yeet the Bear if 100 Acre Wood is Junk or Vanilla"
                )

        if self.proofDepth == locationDepth.Superbosses:
            if (
                locationType.AS not in self.enabledLocations
                and locationType.DataOrg not in self.enabledLocations
            ):
                msg = "Either Absent Silhouettes or Data Organization need to be enabled for superboss-only proof depth"
                raise SettingsException(msg)

    def excluded_levels(self) -> list[int]:
        max_level = self.max_level_checks
        if max_level == 99:
            return self.excludeFrom99
        elif max_level == 50:
            return self.excludeFrom50
        elif max_level == 1:
            return list(range(1, 100))
        else:
            raise SettingsException(f"Incorrect level choice {max_level}")
        
    def double_stat_levels(self) -> list[int]:
        return list(range(self.min_double_stat_level, self.max_double_stat_level))

    def sora_exp(self) -> list[int]:
        return experienceValues.get_sora_exp(
            self.max_level_checks, self.sora_exp_multiplier, self.sora_exp_curve
        )

    def companion_exp(self) -> list[int]:
        return experienceValues.get_companion_exp(
            self.sora_exp_multiplier
        )

    def valor_exp(self) -> list[int]:
        return experienceValues.get_form_exp(
            form.ValorForm, self.valor_exp_multiplier, self.valor_exp_curve
        )

    def wisdom_exp(self) -> list[int]:
        return experienceValues.get_form_exp(
            form.WisdomForm, self.wisdom_exp_multiplier, self.wisdom_exp_curve
        )

    def limit_exp(self) -> list[int]:
        return experienceValues.get_form_exp(
            form.LimitForm, self.limit_exp_multiplier, self.limit_exp_curve
        )

    def master_exp(self) -> list[int]:
        return experienceValues.get_form_exp(
            form.MasterForm, self.master_exp_multiplier, self.master_exp_curve
        )

    def final_exp(self) -> list[int]:
        return experienceValues.get_form_exp(
            form.FinalForm, self.final_exp_multiplier, self.final_exp_curve
        )

    def summon_exp(self) -> list[int]:
        return experienceValues.get_summon_exp(
            self.summon_exp_multiplier, self.summon_exp_curve
        )



from dataclasses import dataclass

from List.configDict import itemType
from List.inventory.item import InventoryItem


@dataclass(frozen=True)
class BonusItem(InventoryItem):
    id: int
    name: str
    type: itemType


ItemSlotUp = BonusItem(463, "Item Slot Up", itemType.SLOT)  # Dummy 16
MaxHpUp = BonusItem(470, "Max HP Up", itemType.GAUGE)  # Dummy 23
MaxMpUp = BonusItem(471, "Max MP Up", itemType.GAUGE)  # Dummy 24
DriveGaugeUp = BonusItem(472, "Drive Gauge Up", itemType.GAUGE)  # Dummy 25
ArmorSlotUp = BonusItem(473, "Armor Slot Up", itemType.SLOT)  # Dummy 26
AccessorySlotUp = BonusItem(474, "Accessory Slot Up", itemType.SLOT)  # Dummy 27



from enum import Enum

VANILLA = "vanilla"
RANDOMIZE_ONE = "rand1"
RANDOMIZE_ALL = "randAll"
RANDOMIZE_IN_GAME_ONLY = "randInGameOnly"
RANDOMIZE_CUSTOM_ONLY = "randCustomOnly"


class locationType(str, Enum):
    LoD = "Land of Dragons"
    BC = "Beast's Castle"
    HB = "Hollow Bastion"
    CoR = "Cavern of Remembrance"
    TT = "Twilight Town"
    TWTNW = "The World That Never Was"
    SP = "Space Paranoids"
    Atlantica = "Atlantica"
    PR = "Port Royal"
    OC = "Olympus Coliseum"
    OCCups = "Olympus Cups"
    OCParadoxCup = "Hades Paradox Cup"
    Agrabah = "Agrabah"
    HT = "Halloween Town"
    PL = "Pride Lands"
    DC = "Disney Castle / Timeless River"
    HUNDREDAW = "Hundred Acre Wood"
    STT = "Simulated Twilight Town"
    AS = "Absent Silhouettes"
    Sephi = "Sephiroth"
    LW = "Lingering Will (Terra)"
    Mush13 = "Mushroom 13"
    DataOrg = "Data Organization XIII"
    FormLevel = "Form Levels"
    Level = "Level"
    FormLevel1 = "Level1Form"
    SummonLevel = "SummonLevel"
    Free = "Garden of Assemblage"
    Critical = "Critical Bonuses"
    Puzzle = "Puzzle"
    WeaponSlot = "Slot"
    TTR = "Transport to Remembrance"
    SYNTH = "Synthesis"
    SHOP = "Shop"
    Creations = "Creations"


class locationCategory(str, Enum):
    CHEST = "Chest"
    POPUP = "Popup"
    CREATION = "Creation"
    ITEMBONUS = "Item Bonus"
    STATBONUS = "Stat Bonus"
    HYBRIDBONUS = "Item and Stat Bonus"
    DOUBLEBONUS = "Double Stat Bonus"
    LEVEL = "Level"
    SUMMONLEVEL = "Summon Level"
    VALORLEVEL = "Valor Level"
    WISDOMLEVEL = "Wisdom Level"
    LIMITLEVEL = "Limit Level"
    MASTERLEVEL = "Master Level"
    FINALLEVEL = "Final Level"
    WEAPONSLOT = "Weapon Slot"

    @staticmethod
    def bonus_categories():
        return [
            locationCategory.DOUBLEBONUS,
            locationCategory.HYBRIDBONUS,
            locationCategory.ITEMBONUS,
            locationCategory.STATBONUS,
        ]


class itemDifficulty(str, Enum):
    SUPEREASY = "Super Easy"
    EASY = "Easy"
    SLIGHTLY_EASY = "Slightly Easy"
    NORMAL = "Normal"
    SLIGHTLY_HARD = "Slightly Hard"
    HARD = "Hard"
    VERYHARD = "Very Hard"
    INSANE = "Insane"
    NIGHTMARE = "Nightmare"

    
class itemBias(str, Enum):
    VERY_EARLY = "Very Early"
    EARLY = "Early"
    SLIGHTLY_EARLY = "Slightly Early"
    NOBIAS = "No Bias"
    SLIGHTLY_LATE = "Slightly Late"
    LATE = "Late"
    VERY_LATE = "Very Late"
    SUPER_LATE = "Very Very Late"
    NIGHTMARE = "As Late as Possible"


class locationDepth(str, Enum):
    Anywhere = "Anywhere"
    NonSuperboss = "SecondVisit"  # Keep an old naming for compatibility
    FirstVisit = "FirstVisit"
    FirstBoss = "FirstBoss"
    SecondVisitOnly = "SecondVisitOnly"
    LastStoryBoss = "SecondBoss"  # Keep an old naming for compatibility
    Superbosses = "DataFight"  # Keep an old naming for compatibility
    NoFirstVisit = "NoFirstVisit"


def location_depth_choices() -> dict[locationDepth, str]:
    return {
        locationDepth.Anywhere: "Anywhere",
        locationDepth.NonSuperboss: "Non-Superboss",
        locationDepth.FirstVisit: "First Visit",
        locationDepth.FirstBoss: "First Visit Boss",
        locationDepth.SecondVisitOnly: "Second Visit",
        locationDepth.LastStoryBoss: "Last Story Boss",
        locationDepth.Superbosses: "Superbosses",
        locationDepth.NoFirstVisit: "Non First Visits",
    }


class expCurve(str, Enum):
    DAWN = "Dawn"
    MIDDAY = "Midday"
    DUSK = "Dusk"

    @staticmethod
    def from_name(name: str):
        return next(c for c in expCurve if c.name == name)


class itemType(str, Enum):
    PROOF_OF_CONNECTION = "Proof of Connection"
    PROOF_OF_PEACE = "Proof of Peace"
    PROOF_OF_NONEXISTENCE = "Proof of Nonexistence"
    PROMISE_CHARM = "Promise Charm"
    FIRE = "Fire"
    BLIZZARD = "Blizzard"
    THUNDER = "Thunder"
    CURE = "Cure"
    MAGNET = "Magnet"
    REFLECT = "Reflect"
    GROWTH_ABILITY = "Growth Ability"
    ACTION_ABILITY = "Action Ability"
    SUPPORT_ABILITY = "Support Ability"
    TORN_PAGE = "Torn Page"
    KEYBLADE = "Keyblade"
    STAFF = "Staff"
    DONALD_ABILITY = "Donald Ability"
    GOOFY_ABILITY = "Goofy Ability"
    SHIELD = "Shield"
    ARMOR = "Armor"
    ACCESSORY = "Accessory"
    ITEM = "Item"
    FORM = "Form"
    MAP = "Map"
    RECIPE = "Recipe"
    SUMMON = "Summon"
    REPORT = "Report"
    KEYITEM = "KeyItem"
    STORYUNLOCK = "Story Unlock"
    MUNNY_POUCH = "Munny Pouch"
    MANUFACTORYUNLOCK = "Manufactory Unlock"
    TROPHY = "OC Trophy"
    OCSTONE = "Olympus Stone"
    GAUGE = "Gauge Up"
    SLOT = "Slot Up"
    SYNTH = "SYNTH"
    MULTIWORLD = "MultiWorld"
    PIECE="Puzzle Piece"
    OBJECTIVE="Objective"


class itemRarity(str, Enum):
    COMMON = "Common"
    UNCOMMON = "Uncommon"
    RARE = "Rare"
    MYTHIC = "Mythic"


class SoraLevelOption(str, Enum):
    LEVEL_1 = "Level"
    LEVEL_50 = "ExcludeFrom50"
    LEVEL_99 = "ExcludeFrom99"


class StartingMovementOption(str, Enum):
    DISABLED = "Disabled"
    RANDOM_3 = "3Random"
    RANDOM_5 = "Random"
    RANDOM_7 = "7Random"
    RANDOM_9 = "9Random"
    LEVEL_1 = "Level_1"
    LEVEL_2 = "Level_2"
    LEVEL_3 = "Level_3"
    LEVEL_4 = "Level_4"


class AbilityPoolOption(str, Enum):
    DEFAULT = "default"
    RANDOMIZE = "randomize"
    RANDOMIZE_SUPPORT = "randomize support"
    RANDOMIZE_STACKABLE = "randomize stackable"


class ItemAccessibilityOption(str, Enum):
    ALL = "all"
    BEATABLE = "beatable"


class SoftlockPreventionOption(str, Enum):
    DEFAULT = "default"
    REVERSE = "reverse"
    BOTH = "both"


class DisableFinalOption(str, Enum):
    DEFAULT = "default"
    NO_ANTIFORM = "no_antiform"
    NO_FINAL = "no_final"


class BattleLevelOption(str, Enum):
    NORMAL = "Normal"
    SHUFFLE = "Shuffle"
    OFFSET = "Offset"
    RANDOM_WITHIN_RANGE = "Within Range of Normal"
    RANDOM_MAX_50 = "Random"  # Old enum naming kept for compatibility
    SCALE_TO_50 = "Scale to 50"


class ObjectivePoolOption(str, Enum):
    ALL = "All Objectives"
    BOSSES = "Bosses Only"
    LASTSTORY = "Last Story Check"
    NOBOSSES = "Everything but Bosses"
    HITLIST = "Spike Hit List"


class LevelUpStatBonus(str, Enum):
    STRENGTH = "Strength"
    MAGIC = "Magic"
    DEFENSE = "Defense"
    AP = "AP"


class HintType(str, Enum):
    DISABLED = "Disabled"
    JSMARTEE = "JSmartee"
    SHANANAS = "Shananas"
    PATH = "Path"
    POINTS = "Points"
    SPOILER = "Spoiler"


class StartingVisitMode(str, Enum):
    ALL = "All Visits"
    FIRST = "First Visits"
    NONE = "No Visits"
    RANDOM = "Random Visits"
    SPECIFIC = "Specific Visits"
    CUSTOM = "Custom"


class FinalDoorRequirement(str, Enum):
    THREE_PROOF = "Three Proofs"
    OBJECTIVES = "Objectives"
    EMBLEMS = "Emblems"



from enum import Enum

from List.configDict import locationType
from List.inventory import ability, misc
from List.location.graph import DefaultLogicGraph, RequirementEdge, chest, popup, stat_bonus, item_bonus, \
    LocationGraphBuilder, START_NODE
from Module.itemPlacementRestriction import ItemPlacementHelpers


class NodeId(str, Enum):
    TwilightTownMapPopup = "Twilight Town Map Popup"
    MunnyPouchPopup = "Munny Pouch Popup"
    RoxasStation = "Roxas Station"
    RoxasStationChests = "Roxas Station Chests"
    TwilightThorn = "Twilight Thorn"
    Axel1 = "Axel 1"
    StruggleChampion = "Struggle Champion"
    SimulatedCentralStation = "STT Central Station"
    SimulatedCentralStationChests = "STT Central Station Chests"
    SimulatedSunsetTerrace = "STT Sunset Terrace"
    SimulatedSunsetTerraceChests = "STT Sunset Terrace Chests"
    SimulatedMansionFoyer = "STT Mansion Foyer"
    SimulatedMansionFoyerChests = "STT Mansion Foyer Chests"
    SimulatedMansionDiningRoom = "STT Mansion Dining Room"
    SimulatedMansionDiningRoomChests = "STT Mansion Dining Room Chests"
    NaminesRoom = "Namine's Room"
    SimulatedMansionLibrary = "STT Mansion Library"
    SimulatedMansionLibraryChests = "STT Mansion Library Chests"
    Axel2 = "Axel 2"
    SimulatedMansionBasement = "STT Mansion Basement"
    SimulatedMansionBasementChests = "STT Mansion Basement Chests"
    DataRoxas = "Data Roxas"


class CheckLocation(str, Enum):
    TwilightTownMap = "Twilight Town Map"
    MunnyPouchOlette = "Munny Pouch (Olette)"
    StationDusks = "Station Dusks"
    StationOfSerenityPotion = "Station of Serenity Potion"
    StationOfCallingPotion = "Station of Calling Potion"
    TwilightThorn = "Twilight Thorn"
    Axel1 = "Axel 1"
    StruggleWinnerChampionBelt = "(Struggle Winner) Champion Belt"
    StruggleLoserMedal = "(Struggle Loser) Medal"
    StruggleTrophy = "The Struggle Trophy"
    CentralStationPotion1 = "Central Station Potion (1)"
    CentralStationHiPotion = "STT Central Station Hi-Potion"
    CentralStationPotion2 = "Central Station Potion (2)"
    SunsetTerraceAbilityRing = "Sunset Terrace Ability Ring"
    SunsetTerraceHiPotion = "Sunset Terrace Hi-Potion"
    SunsetTerracePotion1 = "Sunset Terrace Potion (1)"
    SunsetTerracePotion2 = "Sunset Terrace Potion (2)"
    MansionFoyerHiPotion = "Mansion Foyer Hi-Potion"
    MansionFoyerPotion1 = "Mansion Foyer Potion (1)"
    MansionFoyerPotion2 = "Mansion Foyer Potion (2)"
    MansionDiningRoomElvenBandanna = "Mansion Dining Room Elven Bandanna"
    MansionDiningRoomPotion = "Mansion Dining Room Potion"
    NaminesSketches = "Naminé´s Sketches"
    MansionMap = "Mansion Map"
    MansionLibraryHiPotion = "Mansion Library Hi-Potion"
    Axel2 = "Axel 2"
    MansionBasementCorridorHiPotion = "Mansion Basement Corridor Hi-Potion"
    DataRoxasMagicBoost = "Roxas (Data) Magic Boost"


class STTLogicGraph(DefaultLogicGraph):
    def __init__(self,reverse_rando,keyblade_unlocks):
        DefaultLogicGraph.__init__(self,NodeId)
        keyblade_lambda = lambda inv : not keyblade_unlocks or ItemPlacementHelpers.need_stt_keyblade(inv)
        self.logic[NodeId.RoxasStation][NodeId.RoxasStationChests] = keyblade_lambda
        self.logic[NodeId.SimulatedCentralStation][NodeId.SimulatedCentralStationChests] = keyblade_lambda
        self.logic[NodeId.SimulatedSunsetTerrace][NodeId.SimulatedSunsetTerraceChests] = keyblade_lambda
        self.logic[NodeId.SimulatedMansionFoyer][NodeId.SimulatedMansionFoyerChests] = keyblade_lambda
        self.logic[NodeId.SimulatedMansionDiningRoom][NodeId.SimulatedMansionDiningRoomChests] = keyblade_lambda
        self.logic[NodeId.SimulatedMansionLibrary][NodeId.SimulatedMansionLibraryChests] = keyblade_lambda
        self.logic[NodeId.SimulatedMansionBasement][NodeId.SimulatedMansionBasementChests] = keyblade_lambda
        if not reverse_rando:
            self.logic[START_NODE][NodeId.TwilightTownMapPopup] = ItemPlacementHelpers.stt_check
        else:
            self.logic[START_NODE][NodeId.SimulatedMansionFoyer] = ItemPlacementHelpers.stt_check

def make_graph(graph: LocationGraphBuilder):
    stt = locationType.STT
    stt_logic = STTLogicGraph(graph.reverse_rando,graph.keyblades_unlock_chests)
    graph.add_logic(stt_logic)

    twilight_town_map_popup = graph.add_location(NodeId.TwilightTownMapPopup, [
        popup(319, CheckLocation.TwilightTownMap, stt),
    ])
    munny_pouch_popup = graph.add_location(NodeId.MunnyPouchPopup, [
        popup(288, CheckLocation.MunnyPouchOlette, stt, vanilla=misc.MunnyPouchOlette),
    ])
    roxas_station_chests = graph.add_location(NodeId.RoxasStationChests, [
        chest(315, CheckLocation.StationOfSerenityPotion, stt),
        chest(472, CheckLocation.StationOfCallingPotion, stt),
    ])
    roxas_station = graph.add_location(NodeId.RoxasStation, [
        item_bonus(54, CheckLocation.StationDusks, stt, vanilla=ability.AerialRecovery),
    ])
    twilight_thorn = graph.add_location(NodeId.TwilightThorn, [
        item_bonus(33, CheckLocation.TwilightThorn, stt, vanilla=ability.Guard),
    ])
    axel_1 = graph.add_location(NodeId.Axel1, [
        item_bonus(73, CheckLocation.Axel1, stt, vanilla=ability.Scan),
    ])
    struggle_champion = graph.add_location(NodeId.StruggleChampion, [
        popup(389, CheckLocation.StruggleWinnerChampionBelt, stt),
        popup(390, CheckLocation.StruggleLoserMedal, stt),
        popup(519, CheckLocation.StruggleTrophy, stt, vanilla=misc.StruggleTrophy),
    ])
    central_station_chests = graph.add_location(NodeId.SimulatedCentralStationChests, [
        chest(428, CheckLocation.CentralStationPotion1, stt),
        chest(429, CheckLocation.CentralStationHiPotion, stt),
        chest(430, CheckLocation.CentralStationPotion2, stt),
    ])
    central_station = graph.add_location(NodeId.SimulatedCentralStation, [])
    sunset_terrace_chests = graph.add_location(NodeId.SimulatedSunsetTerraceChests, [
        chest(434, CheckLocation.SunsetTerraceAbilityRing, stt),
        chest(435, CheckLocation.SunsetTerraceHiPotion, stt),
        chest(436, CheckLocation.SunsetTerracePotion1, stt),
        chest(437, CheckLocation.SunsetTerracePotion2, stt),
    ])
    sunset_terrace = graph.add_location(NodeId.SimulatedSunsetTerrace, [])
    mansion_foyer_chests = graph.add_location(NodeId.SimulatedMansionFoyerChests, [
        chest(449, CheckLocation.MansionFoyerHiPotion, stt),
        chest(450, CheckLocation.MansionFoyerPotion1, stt),
        chest(451, CheckLocation.MansionFoyerPotion2, stt),
    ])
    mansion_foyer = graph.add_location(NodeId.SimulatedMansionFoyer, [])
    mansion_dining_room_chests = graph.add_location(NodeId.SimulatedMansionDiningRoomChests, [
        chest(455, CheckLocation.MansionDiningRoomElvenBandanna, stt),
        chest(456, CheckLocation.MansionDiningRoomPotion, stt),
    ])
    mansion_dining_room = graph.add_location(NodeId.SimulatedMansionDiningRoom, [])
    namines_room = graph.add_location(NodeId.NaminesRoom, [
        popup(289, CheckLocation.NaminesSketches, stt),
        popup(483, CheckLocation.MansionMap, stt),
    ])
    mansion_library_chests = graph.add_location(NodeId.SimulatedMansionLibraryChests, [
        chest(459, CheckLocation.MansionLibraryHiPotion, stt),
    ])
    mansion_library = graph.add_location(NodeId.SimulatedMansionLibrary, [])
    axel_2 = graph.add_location(NodeId.Axel2, [
        stat_bonus(34, CheckLocation.Axel2, stt),
    ])
    mansion_basement_chests = graph.add_location(NodeId.SimulatedMansionBasementChests, [
        chest(463, CheckLocation.MansionBasementCorridorHiPotion, stt),
    ])
    mansion_basement = graph.add_location(NodeId.SimulatedMansionBasement, [])

    data_roxas = graph.add_location(NodeId.DataRoxas, [
        popup(558, CheckLocation.DataRoxasMagicBoost, [stt, locationType.DataOrg]),
    ])

    graph.register_superboss(data_roxas)

    graph.add_edge(roxas_station, roxas_station_chests)
    graph.add_edge(central_station, central_station_chests)
    graph.add_edge(sunset_terrace, sunset_terrace_chests)
    graph.add_edge(mansion_foyer, mansion_foyer_chests)
    graph.add_edge(mansion_dining_room, mansion_dining_room_chests)
    graph.add_edge(mansion_library, mansion_library_chests)
    graph.add_edge(mansion_basement, mansion_basement_chests)

    if not graph.reverse_rando:
        graph.add_edge(START_NODE, twilight_town_map_popup)
        graph.add_edge(twilight_town_map_popup, munny_pouch_popup)
        graph.add_edge(munny_pouch_popup, roxas_station)
        graph.add_edge(roxas_station, twilight_thorn, RequirementEdge(battle=True))
        graph.add_edge(twilight_thorn, axel_1, RequirementEdge(battle=True))
        graph.add_edge(axel_1, struggle_champion, RequirementEdge(battle=True))
        graph.add_edge(struggle_champion, central_station)
        graph.add_edge(struggle_champion, sunset_terrace)
        graph.add_edge(sunset_terrace, mansion_foyer, RequirementEdge(battle=True))
        graph.add_edge(mansion_foyer, mansion_dining_room)
        graph.add_edge(mansion_foyer, namines_room)
        graph.add_edge(namines_room, mansion_library)
        graph.add_edge(namines_room, axel_2, RequirementEdge(battle=True))
        graph.add_edge(axel_2, mansion_basement)
        graph.add_edge(mansion_basement, data_roxas, RequirementEdge(battle=True))
        graph.register_first_boss(axel_2)
        graph.register_last_story_boss(axel_2)
    else:
        graph.add_edge(START_NODE, mansion_foyer)
        graph.add_edge(mansion_foyer, mansion_dining_room)
        graph.add_edge(mansion_foyer, namines_room)
        graph.add_edge(namines_room, mansion_library)
        graph.add_edge(namines_room, axel_2, RequirementEdge(battle=True))
        graph.add_edge(axel_2, mansion_basement)
        graph.add_edge(mansion_basement, central_station)
        graph.add_edge(mansion_basement, sunset_terrace)
        graph.add_edge(sunset_terrace, axel_1, RequirementEdge(battle=True))
        graph.add_edge(axel_1, struggle_champion, RequirementEdge(battle=True))
        graph.add_edge(struggle_champion, roxas_station)
        graph.add_edge(roxas_station, twilight_thorn, RequirementEdge(battle=True))
        graph.add_edge(twilight_thorn, twilight_town_map_popup, RequirementEdge(battle=True))
        graph.add_edge(twilight_town_map_popup, munny_pouch_popup)
        graph.add_edge(munny_pouch_popup, data_roxas, RequirementEdge(battle=True))
        graph.register_first_boss(munny_pouch_popup)
        graph.register_last_story_boss(munny_pouch_popup)
